{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    2006 - 2009

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 05.01.2009

  --------------------------------------------------------------------------- }


{ TSqlitePassDatabase }

Constructor TSqlitePassDatabase.Create
(AOwner : TComponent);
begin
  Inherited Create(AOwner);
  {$IFDEF DEBUG}
  Debugger := TSqlitePassDebugger.Create;
  {$ENDIF}
  { Events }
  FBeforeConnect      := Nil;
  FAfterConnect       := Nil;
  FBeforeDisconnect   := Nil;
  FAfterDisconnect    := Nil;
  FDataTypeConversion := Nil;
  { Properties }
  FState := [];
  FConnected := False;
  FReadOnly := False;
  FShowSysObjects := False;
  FSmartOpen := True;
  FDatabaseType := dbtUnknown;
  FQueryTimeout := 0;
  FDatabaseError := TSqlitePassDatabaseError.Create(Self);
  { First, creates the appropriate DbEngine }
  FEngine := TSqlitePassEngine.Create(Self);
  { Must set FTranslator to Nil }
  FTranslator := Nil;
  FVersionInfo := TSqlitePassDatabaseVersionInfo.Create(Self);
  FOptions := TSqlitePassDatabaseOptions.Create(Self);
  FDataTypeOptions := TSqlitePassDatabaseDataTypeOptions.Create(Self);
  FTableDefs := TSqlitePassTableDefs.Create(Self);
  FQueryDefs := TSqlitePassQueryDefs.Create(Self);
  FIndexDefs := TSqlitePassDatabaseIndexDefs.Create(Self);
  FViews := TSqlitePassViews.Create(Self);
  FTriggers := TSqlitePassTriggers.Create(Self);
  FSQLStmtDefs := TSqlitePassSQLStmtDefs.Create(Self);
  FDatasets := TSqlitePassDatasets.create(Self);
  FDatabases := TSqlitePassDatabasesAttached.Create(Self);
end;

Destructor  TSqlitePassDatabase.Destroy;
begin
   Connected := False;
   If Assigned(FTranslator)
      then FTranslator.Free;
   FDatabases.Free;
   FDatasets.Free;
   FSQLStmtDefs.Free;
   FTriggers.Free;
   FViews.Free;
   FIndexDefs.Free;
   FQueryDefs.Free;
   FTableDefs.Free;
   FDataTypeOptions.Free;
   FOptions.Free;
   FVersionInfo.Free;
   FEngine.Free;
   FDatabaseError.Free;
  {$IFDEF DEBUG}
  Debugger.Free;
  {$ENDIF}
 Inherited Destroy;
end;

Procedure TSqlitePassDatabase.SetFConnected(Value : Boolean);
begin
if (Value <> FConnected) and not (csloading in ComponentState) then
   begin
   If Value
      then Open
      else Close;
   end;
end;

function TSqlitePassDatabase.CheckCanOpen(VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow): Boolean;
var
DatabasePath: String;

begin
Result := True;
Try
if FConnected then Exit;

DatabasePath := FDatabase;
  if (DatabasePath='')
     then DatabaseError.RaiseException(Msg1005, -1, Self, VerboseLevel);

  { Gets the database fullpath if FSmartOpen is True }
  if FSmartOpen and (not FileExists(DatabasePath))
    then if csDesigning in ComponentState
            then DatabasePath := GetWorkDir+ExtractFileName(FDatabase)
            else DatabasePath := GetExeDir+ExtractFileName(FDatabase);

  if not FileExists(DatabasePath)
     then DatabaseError.RaiseException(Msg1038+CRLF+DatabasePath, -1, Self, VerboseLevel);

  FEngine.QueryTimeout := FQueryTimeOut;
  FCollatingOrder := GetFCollatingOrder;

  { Tries to open the database and stores the handle to the database }
  Engine.OpenDatabase(DatabasePath, FSqliteLibrary);
  Include(FState,dbsInternalOpen);
Except
  Result := False;
end;
end;

Procedure TSqlitePassDatabase.Open;
begin
if FConnected then Exit;
if Assigned(FBeforeConnect) then FBeforeConnect(Self);
Try
If CheckCanOpen then
   begin
   FConnected:=True;
   RefreshDefinitions;
   Idle;
   if Assigned(FAfterConnect) then FAfterConnect(Self);
   end;
Except
 Close;
 DatabaseError.RaiseException(Msg1007);
end;
end;

Procedure TSqlitePassDatabase.Close;
begin
if FConnected then
   begin
     Try
     if Assigned(FBeforeDisconnect) then FBeforeDisconnect(Self);
     { We save DatatypeOptions to Database if required }
     if (FDatatypeOptions.SaveOptions <> []) and not (soManual in FDatatypeOptions.SaveOptions)
        then FDatatypeOptions.SaveToDatabase;
     FDatasets.Close;
     { We detach previously attached databases first }
     Databases.DetachAll;
     { Clear the Error Log }
     FDatabaseError.ErrorList.Clear;
     FEngine.CloseDatabase;
     if Assigned(FAfterDisconnect) then FAfterDisconnect(Self);
     finally
     FConnected:=False;
     end;
   end;
end;

procedure TSqlitePassDatabase.OpenAttachedDatabases;
var
i: Integer;
NewDb: TSqlitePassDatabase;
begin

{ Check if the Attached Databases List is already in use }
if dbsRefreshingDefinitions in FState
   then Exit;

FDatabasesList := TList.Create;
{ Create temporary databases to get informations from attached Databases }
for i := 0 to Pred(FDatabases.Count) do
   begin
   NewDb := TSqlitePassDatabase.Create(nil);
   NewDb.Database := FDatabases[i].FDatabase;
   NewDb.Open;
   FDatabasesList.Add(NewDb);
   end;
end;

procedure TSqlitePassDatabase.CloseAttachedDatabases;
var
i: Integer;
begin
{ Check if the Attached Databases List need to be used again }
if dbsRefreshingDefinitions in FState
   then Exit;

for i := 0 to Pred(FDatabases.Count)
   do TSqlitePassDatabase(FDatabasesList[i]).Free;

FDatabasesList.Free;
end;

{ Gets information about the database tables, queries...
  including attached databases info }
Procedure TSqlitePassDatabase.RefreshDefinitions;
begin
 if Not CheckCanOpen(vlSilent) then Exit;
 try
 OpenAttachedDatabases;
 Include(FState, dbsRefreshingDefinitions);
 TableDefs.Refresh;
 QueryDefs.Refresh;
 IndexDefs.Refresh;
 Views.Refresh;
 Triggers.Refresh;
 SQLStmtDefs.Refresh;
 finally
 Exclude(FState, dbsRefreshingDefinitions);
 CloseAttachedDatabases;
 If Not Connected
    then FEngine.CloseDatabase;
 end;
end;

Procedure TSqlitePassDatabase.ClearDefinitions;
begin
  TableDefs.Clear;
  QueryDefs.Clear;
  IndexDefs.Clear;
  Views.Clear;
  Triggers.Clear;
  SQLStmtDefs.Clear;
end;

function TSqlitePassDatabase.GetStrPragma(PragmaName: String): String;
var
Recordset : TSqlitePassRecordset;
begin
if Assigned(FEngine) and FConnected then
   begin
   Recordset := TSqlitePassRecordset.Create(Self);
   With Recordset do
        begin
        Open('PRAGMA ' + PragmaName + ';');
        Result := ItemsTextValue[0, 0];
        Free;
        end;
   end
   else Result := '';
end;

function TSqlitePassDatabase.GetIntPragma(PragmaName: String): Integer;
begin
 Result := StrToIntDef(GetStrPragma(PragmaName),-1);
end;

procedure TSqlitePassDatabase.SetPragma(PragmaName: string; value: Boolean);
var
 s:string;
begin
if Assigned(FEngine) and FConnected then
   begin
   if Value
      then s:='ON'
       else s:='OFF';
   FEngine.ExecSQL(Format('PRAGMA %s=%s',[PragmaName, s]));
   end;
end;

procedure TSqlitePassDatabase.SetPragma(PragmaName: string; value: integer);
begin
if Assigned(FEngine) and FConnected
   then FEngine.ExecSQL(Format('PRAGMA %s=%d',[PragmaName, Value]));
end;

procedure TSqlitePassDatabase.SetPragma(PragmaName, Value: string);
begin
if Assigned(FEngine) and FConnected
   then FEngine.ExecSQL(Format('PRAGMA %s=%s',[PragmaName, Value]));
end;

Procedure TSqlitePassDatabase.Idle;
begin
// TODO : TSqlitePassDatabase.Idle - To be implemented
end;

Procedure TSqlitePassDatabase.SetFDatabaseVersion(Value : String);
begin
  // This property is read only
end;

procedure TSqlitePassDatabase.SetFUser(const Value: Integer);
begin
  if FUser=Value then exit;
  FUser:=Value;
end;

Procedure TSqlitePassDatabase.SetFVersionInfo(Value : String);
begin
  // This property is read only
end;

Function TSqlitePassDatabase.GetFCollatingOrder:String;
begin
Result := '';
// TODO : TSqlitePassDatabase.GetFCollatingOrder - To be implemented
end;

Procedure TSqlitePassDatabase.SetFShowSysObjects(Value : Boolean);
begin
 FShowSysObjects := Value;
end;

Procedure TSqlitePassDatabase.SetFReadOnly(Value : Boolean);
begin
 FReadOnly := Value;
end;

function TSqlitePassDatabase.IsSystemTable(TableName: String): Boolean;
begin
Result := Translator.IsSystemTable(TableName);
end;

function TSqlitePassDatabase.Compact: Integer;
var
f: TFileStream;
OriginalSize: Integer;
begin
f := TFileStream.Create(FDatabase, fmOpenRead or fmShareDenyNone);
OriginalSize := f.Size;
Engine.ExecSQL('Vacuum');
Result := OriginalSize - f.Size;
end;

function TSqlitePassDatabase.AttachDatabase(FAttachedDatabase, DatabaseAlias: String): Boolean;
var
AttachedDb: TSqlitePassDatabaseAttached;
ExtPos: Integer;
begin
Result := False;

If FDatabaseType <> DatabaseTypeFromFileName(FAttachedDatabase)
   then DatabaseError.RaiseException(Msg1041 + ' : '+FAttachedDatabase+'.');

if DatabaseAlias = ''
   then begin
        DatabaseAlias := StringReplace(ExtractFileName(FAttachedDatabase), ' ', '', [rfReplaceAll]);
        ExtPos := Pos('.', DatabaseAlias);
        DatabaseAlias := System.Copy(DatabaseAlias, 0, Pred(ExtPos));
        end;

Try
  Engine.Execsql('ATTACH DATABASE ''' + FAttachedDatabase + ''' AS ' + DatabaseAlias + ';');
  AttachedDb := TSqlitePassDatabaseAttached.Create(Databases);
  AttachedDb.Database := FAttachedDatabase;
  AttachedDb.FAliasName := DatabaseAlias;
  FDatabases.Add(AttachedDb);
  Result := True;
Finally
  RefreshDefinitions;
end;
end;

function TSqlitePassDatabase.DetachDatabase(DatabaseAlias: String): Boolean;
var
AttachedDb: TSqlitePassDatabaseAttached;
begin
Result := False;
try
AttachedDb := FDatabases.DatabaseByName(DatabaseAlias);
if Assigned(AttachedDb) then
   begin
     try
      Engine.ExecSQL('DETACH DATABASE ' + DatabaseAlias + ';');
     finally
      FDatabases.Remove(AttachedDb);
      AttachedDb.Free;
      Result := True;
     end;
   end;
finally
 RefreshDefinitions;
end;
end;

procedure TSqlitePassDatabase.SetFDatabaseType(const Value: TSqlitePassDatabaseType);
begin
FDatabaseType := Value;

{ Creates the appropriate FieldTranslator }
If Assigned(FTranslator)
   then FTranslator.Free;

Case FDatabaseType of
  dbtSqlitePass  : FTranslator := TSqlitePassTranslator.Create(Self);
  dbtKexi        : FTranslator := TSqlitePassTranslator_Kexi.Create(Self);
  dbtSqliteAdmin : FTranslator := TSqlitePassTranslator_SqliteAdmin.Create(Self);
  dbtSqlite4Fpc  : FTranslator := TSqlitePassTranslator.Create(Self);
  dbtSqliteExpert: FTranslator := TSqlitePassTranslator_SqliteExpert.Create(Self);
  else             FTranslator := TSqlitePassTranslator.Create(Self);
  end;

//if Not (csLoading in ComponentState)
//   then RefreshDefinitions;
end;

function TSqlitePassDatabase.DatabaseTypeFromFileName(FileName: String): TSqlitePassDatabaseType;
begin
FileName := Lowercase(FileName);
      if (Pos('.kexi',  FileName)) > 0 then Result := dbtKexi
 else if (Pos('.s3db',  FileName)) > 0 then Result := dbtSqliteAdmin
 else if (Pos('.db3',   FileName)) > 0 then Result := dbtSqliteExpert
 else if (Pos('.s3fpc', FileName)) > 0 then Result := dbtSqlite4Fpc
 else if (Pos('.sp3',   FileName)) > 0 then Result := dbtSqlitePass
 else Result := dbtUnknown;
end;

procedure TSqlitePassDatabase.SetFDatabase(Value: String);
begin
  if FConnected
     then DatabaseError.RaiseException(Msg1010);

  FDatabase := Value;

  if FSmartOpen
     then SetFDatabaseType(DatabaseTypeFromFileName(FDatabase));
end;

procedure TSqlitePassDatabase.SetFSQLiteLibrary(Value: String);
begin
if (FConnected)
   then FDatabaseError.RaiseException(Msg1015);
FSQLiteLibrary:=Value;
end;

function TSqlitePassDatabase.GetFEngine: TSqlitePassEngine;
begin
Try
Result := FEngine
except
Result := nil;
FDatabaseError.RaiseException(Msg3014);
end;
end;


function TSqlitePassDatabase.CreateDatabase(DbName: String; DbType: TSqlitePassDatabaseType;
                                            Encoding: TSqlitePassEncoding = UTF_8; PageSize: TSqlitePassPageSize = 4096;
                                            AutoVacuum: TSqlitePassAutoVacuumType = 0): Boolean;
var
EncodingStr: string;
begin
Result := False;
if Not FileExists(DbName) then
   With TSqlitePassEngine.Create(Self) do
        begin
        Try
        OpenDatabase(DbName, FSqliteLibrary);

        Case Encoding of
             UTF_8:    EncodingStr:= '"UTF-8"';
             UTF_16:   EncodingStr:= '"UTF-16"';
             UTF_16le: EncodingStr:= '"UTF-16le"';
             UTF_16be: EncodingStr:= '"UTF-16be"';
             end;

        Case PageSize of
             512         : PageSize := 512;
             513..1024   : PageSize := 1024;
             1025..2048  : PageSize := 2048;
             2049..4096  : PageSize := 4096;
             4097..8192  : PageSize := 8192;
             8193..16384 : PageSize := 16384;
             16385..32768: PageSize := 32768;
             end;

        ExecSQL('PRAGMA encoding = ' + EncodingStr);
        ExecSQL('PRAGMA pagesize = ' + IntToStr(PageSize));
        ExecSQL('PRAGMA Auto_Vacuum = ' + IntToStr(AutoVacuum));

        { Creates system tables for Kexi }
        if DbType = dbtKexi then
           begin
           ExecSQL(KexiDb_CreateSQLStmt);
           ExecSQL(KexiParts_CreateSQLStmt);
           ExecSQL(KexiObjects_CreateSQLStmt);
           ExecSQL(KexiObjectData_CreateSQLStmt);
           ExecSQL(KexiFields_CreateSQLStmt);
           ExecSQL(KexiBlobs_CreateSQLStmt);
           ExecSQL('INSERT INTO "kexi__db" (db_property, db_value) VALUES ("kexidb_major_ver", "1");');
           ExecSQL('INSERT INTO "kexi__db" (db_property, db_value) VALUES ("kexidb_minor_ver", "8");');
           ExecSQL('INSERT INTO "kexi__db" (db_property, db_value) VALUES ("kexiproject_major_ver", "1");');
           ExecSQL('INSERT INTO "kexi__db" (db_property, db_value) VALUES ("kexiproject_minor_ver", "0");');
           ExecSQL('INSERT INTO "kexi__db" (db_property, db_value) VALUES ("project_caption", "");');
           ExecSQL('INSERT INTO "kexi__db" (db_property, db_value) VALUES ("project_desc", "");');
           end;
        Result := True;
        finally
        CloseDatabase;
        Free;
        end;
        end
end;

Function TSqlitePassDatabase.DeleteDatabase(Const DbName: String): Boolean;
begin
Result := FileExists(DbName);
if Result then
   begin
   Result := SysUtils.DeleteFile(DbName);
   if Not Result then DatabaseError.RaiseExceptionFmt(Msg1046,[DbName]);
   end;
end;

Procedure TSqlitePassDatabase.Loaded;
begin
try
  inherited Loaded;
  RefreshDefinitions;
except
  Application.HandleException(Self)
end;
end;


{ TSqlitePassDatabaseVersionInfo }

constructor TSqlitePassDatabaseVersionInfo.Create(Database: TSqlitePassDatabase);
begin
inherited Create();
FDatabase  := Database;
FComponent := SqlitePassDatabaseVersion;
FPackage   := SqlitePassPackageVersion;
FSchema  := 0;
FDbVersion := '';
GetFSqliteEngine;
end;

destructor TSqlitePassDatabaseVersionInfo.Destroy;
begin
FDatabase := nil;
inherited Destroy;
end;


function TSqlitePassDatabaseVersionInfo.GetFSqliteEngine: String;
begin
if Assigned(FDatabase.Engine) and FDatabase.FConnected
   then Result := SqliteDbv3_SqliteLibVersion
   else Result := '';
end;

function TSqlitePassDatabaseVersionInfo.GetFUserTag: Integer;
begin
 Result := FDatabase.GetIntPragma('user_version');
end;

function TSqlitePassDatabaseVersionInfo.GetFDbSchema: Integer;
begin
 Result := FDatabase.GetIntPragma('schema_version');
end;

procedure TSqlitePassDatabaseVersionInfo.SetFDbVersion(const Value: String);
begin
{ Read only - Used to display infos in object inspector }
end;

procedure TSqlitePassDatabaseVersionInfo.SetFPackage(const Value: String);
begin
{ Read only - Used to display infos in object inspector }
end;

procedure TSqlitePassDatabaseVersionInfo.SetFDbSchema(const Value: Integer);
begin
{ Read only - Used to display infos in object inspector }
end;

procedure TSqlitePassDatabaseVersionInfo.SetFSqliteEngine(const Value: String);
begin
{ Read only - Used to display infos in object inspector }
end;

procedure TSqlitePassDatabaseVersionInfo.SetFUserTag(const Value: Integer);
begin
  FDatabase.SetPragma('user_version', Value);
  FUserTag := Value;
end;

procedure TSqlitePassDatabaseVersionInfo.SetFComponent(
  const Value: String);
begin
{ Read only - Used to display infos in object inspector }
end;

{ TSqlitePassDatabaseOptions }

constructor TSqlitePassDatabaseOptions.Create(Database: TSqlitePassDatabase);
begin
inherited Create;
FDatabase := Database;
CaseSensitiveLike := True;
FLogErrors := False;
QuoteStyle := qsDoubleQuote;
end;

Destructor TSqlitePassDatabaseOptions.Destroy;
begin
FDatabase := nil;
inherited Destroy;
end;

function TSqlitePassDatabaseOptions.GetFAutoVacuum: TSqlitePassAutoVacuumType;
begin
 Result := TSqlitePassAutoVacuumType(FDatabase.GetIntPragma('auto_vacuum'))
end;

function TSqlitePassDatabaseOptions.GetFCacheSize: Integer;
begin
 Result := FDatabase.GetIntPragma('cache_size')
end;

function TSqlitePassDatabaseOptions.GetFCountChanges: Boolean;
begin
 { do Not Use Result := Boolean(FDatabase.GetIntPragma('count_changes'));
   Compiler doesn't warn but Boolean Transtypage is not safe }
 Result := FDatabase.GetIntPragma('count_changes') <> 0;
end;

function TSqlitePassDatabaseOptions.GetFDefaultCacheSize: Integer;
begin
  Result := FDatabase.GetIntPragma('default_cache_size');
end;

function TSqlitePassDatabaseOptions.GetFEmptyResultCallBack: Boolean;
begin
 Result := FDatabase.GetIntPragma('empty_result_callbacks') <> 0;
end;

function TSqlitePassDatabaseOptions.GetFEncoding: TSqlitePassEncoding;
var
s: String;
begin
s := FDatabase.GetStrPragma('encoding');
     if s = 'UTF-8'    then Result := UTF_8
else if s = 'UTF-16'   then Result := UTF_16
else if s = 'UTF-16le' then Result := UTF_16le
else Result := UTF_16be;
end;

function TSqlitePassDatabaseOptions.GetFFullColumnNames: Boolean;
begin
  Result := FDatabase.GetIntPragma('full_column_names') <> 0;
end;

function TSqlitePassDatabaseOptions.GetFPageSize: TSqlitePassPageSize;
begin
  Result := FDatabase.GetIntPragma('page_size');
end;

function TSqlitePassDatabaseOptions.GetFSynchronous: TSqlitePassSynchronous;
begin
  Result := FSynchronous;//TSqlitePassSynchronous(FDatabase.GetIntPragma('synchronous'));
end;

function TSqlitePassDatabaseOptions.GetFTemporaryStorage: TSqlitePassTempStore;
begin
  Result := FTemporaryStorage;//TSqlitePassTempStore(FDatabase.GetIntPragma('temp_store'));
end;

procedure TSqlitePassDatabaseOptions.SetFAutoVacuum(const Value: TSqlitePassAutoVacuumType);
begin
 FDatabase.SetPragma('auto_vacuum',Value);
 FAutoVacuum := Value;
end;

procedure TSqlitePassDatabaseOptions.SetFCacheSize(const Value: Integer);
begin
 FDatabase.SetPragma('cache_size',Value);
 FCacheSize := Value;
end;

procedure TSqlitePassDatabaseOptions.SetFCaseSensitiveLike(const Value: Boolean);
begin
 If Value <> FCaseSensitiveLike then
    begin
    FDatabase.SetPragma('case_sensitive_like',Value);
    FCaseSensitiveLike := Value;
    end;
end;

procedure TSqlitePassDatabaseOptions.SetFCountChanges(const Value: Boolean);
begin
 FDatabase.SetPragma('count_changes',Value);
 FCountChanges := Value;
end;

procedure TSqlitePassDatabaseOptions.SetFDefaultCacheSize(const Value: Integer);
begin
 FDatabase.SetPragma('default_cache_size',Value);
 FDefaultCacheSize := Value;
end;

procedure TSqlitePassDatabaseOptions.SetFEmptyResultCallBack(const Value: Boolean);
begin
 FDatabase.SetPragma('empty_result_callbacks', Value);
 FEmptyResultCallBack := Value;
end;

procedure TSqlitePassDatabaseOptions.SetFEncoding(const Value: TSqlitePassEncoding);
begin
{ Read Only - Used to display info in object inspector
  Dabase encoding can be written once, only when the database is created... }
end;

procedure TSqlitePassDatabaseOptions.SetFFullColumnNames(const Value: Boolean);
begin
 FDatabase.SetPragma('full_column_names', Value);
 FDatabase.SetPragma('short_column_names', not Value);
 FFullColumnNames := Value;
 FShortColumnNames := Not Value;
end;

procedure TSqlitePassDatabaseOptions.SetFPageSize(Value: TSqlitePassPageSize);
begin
// TODO
end;

procedure TSqlitePassDatabaseOptions.SetFSynchronous(const Value: TSqlitePassSynchronous);
begin
 FDatabase.SetPragma('synchronous', Ord(Value));
 FSynchronous := Value;
end;

procedure TSqlitePassDatabaseOptions.SetFTemporaryStorage(const Value: TSqlitePassTempStore);
begin
 FDatabase.SetPragma('temp_store', Ord(Value));
 FTemporaryStorage := Value;
end;

procedure TSqlitePassDatabaseOptions.SetFQuoteStyle(const Value: TSqlitePassQuoteStyle);
begin
 Case Value of
      qsSingleQuote: begin FQuoteOpen := ''''; FQuoteClose := ''''; end;
      qsDoubleQuote: begin FQuoteOpen := '"'; FQuoteClose := '"'; end;
      qsBracket    : begin FQuoteOpen := '['; FQuoteClose := ']'; end;
      else           begin FQuoteOpen := #0; FQuoteClose := #0; end;
      end;
FQuoteStyle := Value;
end;



{ TSqlitePassFieldTypesTranslationRules }


constructor TSqlitePassFieldTypesTranslationRules.Create(
  DataTypeOptions: TSqlitePassDatabaseDataTypeOptions);
begin
inherited Create;
FDataTypeOptions := DataTypeOptions;
end;

Destructor TSqlitePassFieldTypesTranslationRules.Destroy;
begin
ClearAndFreeItems;
FDataTypeOptions := nil;
inherited Destroy;
end;

function TSqlitePassFieldTypesTranslationRules.GetItem(Index: Integer): TSqlitePassFieldTypeTranslationRule;
begin
  Result:=TSqlitePassFieldTypeTranslationRule(Inherited Items[Index]);
end;

procedure TSqlitePassFieldTypesTranslationRules.SetItem(Index: Integer;
  FieldTypeTranslationRule: TSqlitePassFieldTypeTranslationRule);
begin
   Put(Index,FieldTypeTranslationRule);
end;

function TSqlitePassFieldTypesTranslationRules.Add(FieldTypeTranslator: TSqlitePassFieldTypeTranslationRule
  ): Integer;
begin
   Result:=Inherited Add(FieldTypeTranslator);
end;

procedure TSqlitePassFieldTypesTranslationRules.ClearAndFreeItems;
var
i: Integer;
begin
{ Clear the internal items }
for i := 0 to Pred(Count)
    do Items[i].Free;
{ Clear the list pointers }
inherited Clear;
end;

procedure TSqlitePassFieldTypesTranslationRules.SetFieldDataType(FieldDef: TSqlitePassTableFieldDef);

    procedure ConvertNativeFieldTypeToPascalFieldType;
    var
    i: integer;
    FieldTypeName: String;
    begin
    FieldTypeName := LowerCase(FieldDef.NativeDataType);
    For i:=0 to Pred(Count) do
      begin
      Case Items[i].MappingMode of
           mmExact        : if (Items[i].DatatypeName = FieldTypeName)
                               then FieldDef.DataType := Items[i].FieldType;

           mmExactNoCase  : if AnsiCompareText(Items[i].DatatypeName,FieldTypeName) = 0
                               then FieldDef.DataType := Items[i].FieldType;

           mmPartial      : if Pos(Items[i].DatatypeName, FieldTypeName) <> 0
                               then FieldDef.DataType := Items[i].FieldType;

           mmPartialNoCase: if Pos(Uppercase(Items[i].DatatypeName), Uppercase(FieldTypeName)) <> 0
                               then FieldDef.DataType := Items[i].FieldType;

           mmAll          : if (Items[i].DatatypeName = FieldTypeName)
                               then FieldDef.DataType := Items[i].FieldType
                               else if Pos(Items[i].DatatypeName, FieldTypeName) <> 0
                                       then FieldDef.DataType := Items[i].FieldType;
           end;
      if FieldDef.DataType <> ftUnknown
         then Break;
      end;

    { Set a default fieldtype if fieltype is ftUnknown }
    If FieldDef.DataType = ftUnknown
       then FieldDef.DataType := FDataTypeOptions.DefaultFieldType;
    end;


begin { procedure SetFieldDataType }
FieldDef.DataType := ftUnknown;

Case FDataTypeOptions.FDetectionMode of
     dmTypeName  : ConvertNativeFieldTypeToPascalFieldType;
     dmDbSpecific: begin
                   FDataTypeOptions.FDatabase.FTranslator.ConvertNativeFieldTypeToPascalFieldType(FieldDef);
                   if FieldDef.DataType = ftUnknown
                      then ConvertNativeFieldTypeToPascalFieldType;
                   end;
     dmCustom    : begin
                   if Assigned(FDataTypeOptions.FDatabase.OnDataTypeConversion) then
                      begin
                      ConvertNativeFieldTypeToPascalFieldType;
                      if FieldDef.DataType = ftUnknown
                         then FDataTypeOptions.FDatabase.Translator.ConvertNativeFieldTypeToPascalFieldType(FieldDef);
                      FDataTypeOptions.FDatabase.OnDataTypeConversion(FDataTypeOptions.FDatabase, FieldDef.FieldDefs.TableDef, FieldDef);
                      end;
                   end;
     dmForceStr  : FieldDef.DataType := ftString;
     end;
end;


Procedure TSqlitePassFieldTypesTranslationRules.SetDefaultTranslationRules(DatabaseType: TSqlitePassDatabaseType);
Type
TSqlitePassPresetFieldType = record
   Name:string;
   Fieldtype:TFieldType;
   end;

const
{ Default Fields Type when DatabaseType is dtbUnknown }
DefaultFieldsTypes:array[0..21] of TSqlitePassPresetFieldType =
           ((Name:'widestring'; FieldType:ftString), { must be before string }
            (Name:'string'; FieldType:ftString),
            (Name:'smallint'; FieldType:ftSmallInt), { must be before integer }
            (Name:'largeint'; FieldType:ftLargeint), { must be before integer }
            (Name:'integer'; Fieldtype:ftInteger),
            (Name:'word'; FieldType:ftWord),
            (Name:'float'; FieldType:ftFloat),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'curency'; FieldType:ftCurrency),
            (Name:'bcd'; FieldType:ftBCD),
            (Name:'datetime'; FieldType:ftDateTime), { must be before 'date' }
            (Name:'date'; FieldType:ftDate),
            (Name:'time'; FieldType:ftTime),
            (Name:'varbytes'; FieldType:ftVarBytes),
            (Name:'bytes'; FieldType:ftBytes),
            (Name:'autoinc'; FieldType:ftLargeInt),  { Do not set as ftAutoInc }
            (Name:'fmtmemo'; FieldType:ftFmtMemo),   { must be before memo }
            (Name:'memo'; FieldType:ftMemo),
            (Name:'blob'; FieldType:ftBlob),
            (Name:'graphic'; FieldType:ftGraphic),
            (Name:'typedbinary'; FieldType:ftTypedBinary),
            (Name:'fixedchar'; FieldType:ftfixedchar));

{ Default Fields Type when DatabaseType is dtbKexi }
KexiFieldTypes:array[0..12] of TSqlitePassPresetFieldType =
           ((Name:'integer'; Fieldtype:ftInteger),
            (Name:'shortinteger'; FieldType:ftSmallInt),
            (Name:'biginteger'; FieldType:ftLargeint),
            (Name:'byte'; FieldType:ftSmallInt),
            (Name:'double'; FieldType:ftFloat),
            (Name:'float'; FieldType:ftFloat),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'datetime'; FieldType:ftDateTime), { must be before 'date' }
            (Name:'date'; FieldType:ftDate),
            (Name:'time'; FieldType:ftTime),
            (Name:'text'; FieldType:ftString),
            (Name:'clob'; FieldType:ftMemo),
            (Name:'blob'; FieldType:ftBlob));

SqliteAdminFieldTypes:array[0..11] of TSqlitePassPresetFieldType =
           ((Name:'boolean'; FieldType:ftBoolean),
            (Name:'date'; FieldType:ftDate),
            (Name:'float'; FieldType:ftFloat),
            (Name:'integer'; FieldType:ftInteger),
            (Name:'numeric'; FieldType:ftFloat),
            (Name:'real'; FieldType:ftFloat),
            (Name:'text'; FieldType:ftMemo),
            (Name:'timestamp'; FieldType:ftDateTime), { Must be before 'time' }
            (Name:'time'; FieldType:ftTime),
            (Name:'nvarchar'; FieldType:ftString),
            (Name:'varchar'; FieldType:ftString),
            (Name:'blob'; FieldType:ftblob));

{ Default Fields Type when DatabaseType is dbtSqliteExpert }
SqliteExpertFieldTypes:array[0..40] of TSqlitePassPresetFieldType =
           ((Name:'autoinc'; Fieldtype:ftLargeInt), { Do not set as ftAutoinc }
            (Name:'bigint'; FieldType:ftLargeint),
            (Name:'binary'; FieldType:ftBlob),
            (Name:'blob'; FieldType:ftBlob),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'char'; FieldType:ftString),
            (Name:'clob'; FieldType:ftMemo),
            (Name:'currency'; FieldType:ftCurrency),
            (Name:'datetext'; FieldType:ftString),   { must be before 'date' }
            (Name:'datetime'; FieldType:ftDateTime), { must be before 'date' }
            (Name:'date'; FieldType:ftDate),
            (Name:'dec'; FieldType:ftFloat),
            (Name:'decimal'; FieldType:ftFloat),
            (Name:'double precision'; FieldType:ftFloat),
            (Name:'double'; FieldType:ftFloat),
            (Name:'float'; FieldType:ftFloat),
            (Name:'graphic'; FieldType:ftGraphic),
            (Name:'guid'; FieldType:ftMemo), //
            (Name:'int64'; FieldType:ftLargeInt),    { must be before integer }
            (Name:'int'; FieldType:ftInteger),       { must be before integer }
            (Name:'integer'; FieldType:ftInteger),
            (Name:'largeint'; FieldType:ftLargeInt),
            (Name:'memo'; FieldType:ftMemo),
            (Name:'nchar'; FieldType:ftString),
            (Name:'money'; FieldType:ftCurrency),
            (Name:'ntext'; FieldType:ftString),
            (Name:'number'; FieldType:ftFloat),
            (Name:'numeric'; FieldType:ftFloat),
            (Name:'nvarchar2'; FieldType:ftString),
            (Name:'nvarchar'; FieldType:ftString),
            (Name:'raw'; FieldType:ftBlob),
            (Name:'real'; FieldType:ftFloat),
            (Name:'smallint'; FieldType:ftInteger),
            (Name:'smallmoney'; FieldType:ftCurrency),
            (Name:'text'; FieldType:ftString),
            (Name:'timestamp'; FieldType:ftDateTime), { Must be before 'time' }
            (Name:'time'; FieldType:ftTime),
            (Name:'tinyint'; FieldType:ftInteger),
            (Name:'varbinary'; FieldType:ftBlob),
            (Name:'varchar2'; FieldType:ftString),
            (Name:'varchar'; FieldType:ftString));


{ Default Fields Type when DatabaseType is dbtSqlite4Fpc
  Support for TCustomSqliteDataset Component from Luiz Amrico Pereira Cmara}
Sqlite4FpcFieldTypes:array[0..13] of TSqlitePassPresetFieldType =
           ((Name:'autoinc'; Fieldtype:ftLargeInt),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'currency'; FieldType:ftCurrency),
            (Name:'datetime'; FieldType:ftDateTime),  { must be before 'date' }
            (Name:'date'; FieldType:ftDate),            
            (Name:'float'; FieldType:ftFloat),
            (Name:'int'; FieldType:ftInteger),
            (Name:'integer'; FieldType:ftInteger),
            (Name:'largeint'; FieldType:ftLargeInt),
            (Name:'numeric'; FieldType:ftFloat),
            (Name:'text'; FieldType:ftMemo),
            (Name:'time'; FieldType:ftTime),
            (Name:'varchar'; FieldType:ftString),
            (Name:'word'; FieldType:ftWord));

    { Loads the requested Default Datatypes Mapping }
    procedure SetTranslationRules(FieldsTypes: array of TSqlitePassPresetFieldType);
    var
    i: Word;
    FieldTypeTranslator: TSqlitePassFieldTypeTranslationRule;
    begin
    ClearAndFreeItems;
    for i := 0 to High(FieldsTypes) do
       begin
       FieldTypeTranslator := TSqlitePassFieldTypeTranslationRule.Create;
       With FieldTypeTranslator do
            begin
            DatatypeName := FieldsTypes[i].Name;
            MappingMode  := mmAll;
            FieldType    := FieldsTypes[i].FieldType;
            end;
       Add(FieldTypeTranslator);
       end;
    end;

begin
Case DatabaseType of
     dbtKexi         : SetTranslationRules(KexiFieldTypes);
     dbtSqliteExpert : SetTranslationRules(SqliteExpertFieldTypes);
     dbtSqliteAdmin  : SetTranslationRules(SqliteAdminFieldTypes);
     dbtSqlite4Fpc   : SetTranslationRules(Sqlite4FpcFieldTypes);
     else              SetTranslationRules(DefaultFieldsTypes);
     end;
end;


{ TSqlitePassCustomFieldDefs }

constructor TSqlitePassCustomFieldDefs.Create;
begin
inherited Create;
end;

Destructor TSqlitePassCustomFieldDefs.Destroy;
begin
ClearAndFreeItems;
inherited Destroy;
end;

function TSqlitePassCustomFieldDefs.GetItem(Index: Integer): TSqlitePassCustomFieldDef;
begin
  Result:=TSqlitePassCustomFieldDef(Inherited Items[Index]);
end;

procedure TSqlitePassCustomFieldDefs.SetItem(Index: Integer;
  CustomFieldDef: TSqlitePassCustomFieldDef);
begin
   Put(Index,CustomFieldDef);
end;

function TSqlitePassCustomFieldDefs.Add(CustomFieldDef: TSqlitePassCustomFieldDef): Integer;
begin
   Result:=Inherited Add(CustomFieldDef);
end;

procedure TSqlitePassCustomFieldDefs.ClearAndFreeItems;
var
i: Integer;
begin
{ Clear the internal items }
for i := 0 to Pred(Count)
    do Items[i].Free;
{ Clear the list pointers }
inherited Clear;
end;


{ TSqlitePassTranslator }

{ --- constructor / Destructor --- }
constructor TSqlitePassTranslator.Create(Owner: TSqlitePassDatabase);
begin
Inherited Create;
FDatabase := Owner;
FFieldTypesTranslationRules := FDatabase.FDatatypeOptions.TranslationRules;
end;

destructor TSqlitePassTranslator.Destroy;
begin
FFieldTypesTranslationRules := nil;
FDatabase := nil;
Inherited Destroy;
end;


{ ************************************************************************************ }
{ *** Conversion functions from an internal MemRecord storage to a SQLite Db Value *** }

{ Translates a value from internal buffer to an sqlite value. }

{ Boolean }
procedure TSqlitePassTranslator.WordBoolToInteger(Const Value: WordBool; out IntegerValue: Integer);
begin
If Value
   then IntegerValue := 1
   else IntegerValue := 0;
end;

procedure TSqlitePassTranslator.WordBoolToBooleanText(Const Value: WordBool; out StrValue: String);
begin
If Value
   then StrValue := FDatabase.DatatypeOptions.FBooleanTrueValues[0]
   else StrValue := FDatabase.DatatypeOptions.FBooleanFalseValues[0];
end;

{ Date, Time and DateTime Conversion }

procedure TSqlitePassTranslator.IntegerToDateText
         (Const Value: Integer;
          Const DatePattern: String;
          Const YearStart, YearLength, MonthStart, MonthLength, DayStart, DayLength: Word;
          out StrValue: String);
var
Year, Month, Day: Word;
DatePartStr: String;
PStrValue: PChar;
begin
 DecodeDate(Value, Year, Month, Day);
 { Necessary conversion oterwise we would be pointing on FDatabase.DatatypeOptions.FDateFormat }
 StrValue := StrPas(PChar(DatePattern));
 PStrValue := PChar(StrValue);
 {Year}
 DatePartStr := Copy(Format('%0.4d',[Year]), 5-YearLength, YearLength);
 StrMove(PStrValue + Pred(YearStart),PChar(DatePartStr),YearLength);
 {Month}
 DatePartStr := Copy(Format('%0.2d',[Month]), 3-MonthLength, MonthLength);
 StrMove(PStrValue + Pred(MonthStart),PChar(DatePartStr),MonthLength);
 {Day}
 DatePartStr := Copy(Format('%0.2d',[Day]), 3-DayLength, DayLength);
 StrMove(PStrValue + Pred(DayStart),PChar(DatePartStr),DayLength);
end;

procedure TSqlitePassTranslator.IntegerToDateText(Const Value: Integer; out StrValue: String);
begin
 IntegerToDateText(Value,
                   FDatabase.DatatypeOptions.FDateFormat,
                   FDatabase.DatatypeOptions.FDateFormatPattern[0,0],
                   FDatabase.DatatypeOptions.FDateFormatPattern[0,1],
                   FDatabase.DatatypeOptions.FDateFormatPattern[1,0],
                   FDatabase.DatatypeOptions.FDateFormatPattern[1,1],
                   FDatabase.DatatypeOptions.FDateFormatPattern[2,0],
                   FDatabase.DatatypeOptions.FDateFormatPattern[2,1],
                   StrValue);
end;


procedure TSqlitePassTranslator.IntegerToTimeText
         (Const Value: Integer;
          Const TimePattern: String;
          Const HourStart, HourLength, MinStart, MinLength, SecStart, SecLength, MSecStart, MSecLength: Word;
          out StrValue: String);
var
Hour, Min, Sec, MSec: Word;
TimePartStr: String;
PStrValue: PChar;
begin
 DecodeTime(TimeStampToDateTime(MsecsToTimeStamp(Value)), Hour, Min, Sec, MSec);
 { Necessary conversion oterwise we would be pointing on FDatabase.DatatypeOptions.FTimeFormat }
 StrValue := StrPas(PChar(TimePattern));
 PStrValue := PChar(StrValue);
 {Hours}
 TimePartStr := Copy(Format('%0.2d',[Hour]), 3-HourLength, HourLength);
 StrMove(PStrValue + Pred(HourStart),PChar(TimePartStr),HourLength);
 {Minutes}
 TimePartStr := Copy(Format('%0.2d',[Min]), 3-MinLength, MinLength);
 StrMove(PStrValue + Pred(MinStart),PChar(TimePartStr),MinLength);
 {Secondes}
 TimePartStr := Copy(Format('%0.2d',[Sec]), 3-SecLength, SecLength);
 StrMove(PStrValue + Pred(SecStart),PChar(TimePartStr),SecLength);
 {MSecondes are optional, so we check the Time Format}
 if MSecStart > 0 then
    begin
    TimePartStr := Copy(Format('%0.3d',[MSec]), 4-MSecLength, MSecLength);
    StrMove(PStrValue + Pred(MSecStart),PChar(TimePartStr),MSecLength);
    end;
end;

procedure TSqlitePassTranslator.IntegerToTimeText(Const Value: Integer; out StrValue: String);
begin
IntegerToTimeText(Value,
                  FDatabase.DatatypeOptions.FTimeFormat,
                  FDatabase.DatatypeOptions.FTimeFormatPattern[0,0],
                  FDatabase.DatatypeOptions.FTimeFormatPattern[0,1],
                  FDatabase.DatatypeOptions.FTimeFormatPattern[1,0],
                  FDatabase.DatatypeOptions.FTimeFormatPattern[1,1],
                  FDatabase.DatatypeOptions.FTimeFormatPattern[2,0],
                  FDatabase.DatatypeOptions.FTimeFormatPattern[2,1],
                  FDatabase.DatatypeOptions.FTimeFormatPattern[3,0],
                  FDatabase.DatatypeOptions.FTimeFormatPattern[3,1],
                  StrValue);
end;

procedure TSqlitePassTranslator.DoubleToDateTimeText(Const Value: Double; out StrValue: String);
var
TimeStamp: TTimeStamp;
DateStr: String;
begin
 TimeStamp := MSecsToTimeStamp(Value);
 Dec(TimeStamp.Date, DateDelta);
 IntegerToDateText(TimeStamp.Date,
                   FDatabase.DatatypeOptions.FDateTimeFormat,
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[0,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[0,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[1,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[1,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[2,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[2,1],
                   StrValue);

 DateStr := StrValue;
 IntegerToTimeText(TimeStamp.Time,
                   DateStr,
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[3,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[3,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[4,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[4,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[5,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[5,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[6,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[6,1],
                   StrValue);
end;

procedure TSqlitePassTranslator.DoubleToJulianDateTime(Const Value: Double; out DoubleValue: Double);
begin
  DoubleValue := Value - JulianEpoch;
end;

procedure TSqlitePassTranslator.DoubleToUnixDateTime(Const Value: Double; out Int64Value: Int64);
begin
  Int64Value := Trunc(Abs(UnixEpoch - Value)*SecsPerDay);
end;

procedure TSqlitePassTranslator.DoubleToMacDateTime(Const Value: Double; out Int64Value: Int64);
var
  MacEpoch:TDateTime;
begin
  MacEpoch := EncodeDate(1904, 1, 1) + EncodeTime(0, 0, 0, 0);
  Int64Value := Trunc(Abs(MacEpoch-Value)*SecsPerDay);
end;



{ ************************************************************************************** }
{ *** Conversion procedures from a SQLite Db Value to it's internal MemRecord storage *** }

{ Boolean Conversions }

procedure TSqlitePassTranslator.IntegerToWordBool(Const IntegerValue: Integer; out Value: WordBool);
begin
 Value := IntegerValue <> 0;
end;

procedure TSqlitePassTranslator.BooleanTextToWordBool(Const StrValue: String; out Value: WordBool);
var
StrLower: String;
I: Integer;
begin
 StrLower := LowerCase(StrValue);
 For i:=0 to Pred(FDatabase.DatatypeOptions.FBooleanTrueValues.Count) do
     begin
     Value := Pos(FDatabase.DatatypeOptions.FBooleanTrueValues[i], StrLower) > 0;
     if Value then Break;
     end;
end;

{ Date, Time and DateTime Conversion }

procedure TSqlitePassTranslator.DateTextToInteger(Const StrValue: String;
                                                 Const YearStart, YearLength, MonthStart, MonthLength, DayStart, DayLength: Word;
                                                 out Value: Integer);
var
Year, Month, Day: Word;
begin
 Year := StrToIntDef(System.Copy(StrValue,YearStart,YearLength),0);
 Month := StrToIntDef(System.Copy(StrValue,MonthStart,MonthLength),0);
 Day := StrToIntDef(System.Copy(StrValue,DayStart,DayLength),0);
 if (Year>0) and (Year<10000) and (Month in [1..12])
    and (Day>0) and (Day<=MonthDays[IsleapYear(Year),Month]) then
    begin
    Value := Trunc(EncodeDate(Year, Month, Day));
    Inc(Value, DateDelta);
    end;
end;


procedure TSqlitePassTranslator.DateTextToInteger(Const StrValue: String; out Value: Integer);
begin
 DateTextToInteger(StrValue,
                   FDatabase.DatatypeOptions.FDateFormatPattern[0,0],
                   FDatabase.DatatypeOptions.FDateFormatPattern[0,1],
                   FDatabase.DatatypeOptions.FDateFormatPattern[1,0],
                   FDatabase.DatatypeOptions.FDateFormatPattern[1,1],
                   FDatabase.DatatypeOptions.FDateFormatPattern[2,0],
                   FDatabase.DatatypeOptions.FDateFormatPattern[2,1],
                   Value);
end;


procedure TSqlitePassTranslator.TimeTextToInteger(Const StrValue: String;
                               Const HourStart, HourLength, MinStart, MinLength, SecStart, SecLength, MSecStart, MSecLength: Word;
                               out Value: Integer);
var
Hour, Min, Sec, MSec: Word;
begin
 Value := 0;
 Hour := StrToIntDef(System.Copy(StrValue,HourStart,HourLength),99);
 Min  := StrToIntDef(System.Copy(StrValue,MinStart,MinLength),99);
 Sec  := StrToIntDef(System.Copy(StrValue,SecStart,SecLength),99);
 MSec := StrToIntDef(System.Copy(StrValue,MSecStart,MSecLength),0);

 if (Hour in [0..24]) and (Min in [0..60]) and (Sec in [0..60]) and (MSec < 1000)
    then Value := DateTimeToTimeStamp(EncodeTime(Hour, Min, Sec, Msec)).Time;
end;

procedure TSqlitePassTranslator.TimeTextToInteger(Const StrValue: String; out Value: Integer);
begin
 TimeTextToInteger(StrValue,
                   FDatabase.DatatypeOptions.FTimeFormatPattern[0,0],
                   FDatabase.DatatypeOptions.FTimeFormatPattern[0,1],
                   FDatabase.DatatypeOptions.FTimeFormatPattern[1,0],
                   FDatabase.DatatypeOptions.FTimeFormatPattern[1,1],
                   FDatabase.DatatypeOptions.FTimeFormatPattern[2,0],
                   FDatabase.DatatypeOptions.FTimeFormatPattern[2,1],
                   FDatabase.DatatypeOptions.FTimeFormatPattern[3,0],
                   FDatabase.DatatypeOptions.FTimeFormatPattern[3,1],
                   Value);
end;

procedure TSqlitePassTranslator.DateTimeTextToDouble(Const StrValue: String; out Value: Double);
var
TimeStamp: TTimeStamp;
begin
 DateTextToInteger(StrValue,
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[0,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[0,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[1,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[1,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[2,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[2,1],
                   TimeStamp.Date);

 TimeTextToInteger(StrValue,
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[3,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[3,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[4,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[4,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[5,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[5,1],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[6,0],
                   FDatabase.DatatypeOptions.FDateTimeFormatPattern[6,1],
                   TimeStamp.Time);

 Value := TimeStampToMsecs(TimeStamp);
end;

procedure TSqlitePassTranslator.JulianDateTimeToDouble(Const DoubleValue: Double; out Value: Double);
begin
  Value := DoubleValue + JulianEpoch;
end;

procedure TSqlitePassTranslator.UnixDateTimeToDouble(Const Int64Value: Int64; out Value: Double);
begin
  Value := UnixEpoch + Int64Value / SecsPerDay;
end;

procedure TSqlitePassTranslator.MacDateTimeToDouble(Const Int64Value: Int64; out Value: Double);
var
 MacEpoch:TDateTime;
begin
 MacEpoch := EncodeDate(1904, 1, 1) + EncodeTime(0, 0, 0, 0);
 Value := MacEpoch + Int64Value / SecsPerDay;
end;


(*function TSqlitePassTranslator.CreateTable(const Dataset: TSqlitePassDataset): Boolean;
begin
// TODO : TSqlitePassTranslator.CreateTable - To be implemented
end;*)


{ Database objects information is set from sqlite_master and
 sqlite_temp_master tables. They return basic information :
 Rowid, type, name, tbl_name, rootpage, sql (Create stmt)}
procedure TSqlitePassTranslator.GetDatabaseTableDefs(TableDefs: TSqlitePassTableDefs);
const
Sql = 'SELECT Rowid, * FROM sqlite_master WHERE type = "table" ' +
      'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "table" ' +
      'ORDER BY type, tbl_name;';

var
i: Integer;

  Procedure GetTableDefs(AttachedDb: TSqlitePassDatabaseAttached = nil);
  var
  i, NewRecIndex: integer;
  TempDb: TSqlitePassDatabase;
  TableDef: TSqlitePassTableDef;
  TableDefsRecordset: TSqlitePassRecordset;

  begin
  if Assigned(AttachedDb)
     then begin
          TempDb := TSqlitePassDatabase(FDatabase.FDatabasesList[FDatabase.FDatabases.IndexOf(AttachedDb)]);
          TableDefs.AddTableDefs(TempDb.TableDefs, AttachedDb);
          end
     else begin
          TempDb := FDatabase;
          TableDefsRecordset := TSqlitePassRecordset.Create(TempDb);

          try
          TableDefsRecordset.Open(Sql);

          { Result structure is : Type, name, tablename, rootpage, sql }
          {  We manually add the sqlite_master table which is not returned by engine }
          NewRecIndex := TableDefsRecordset.AddRecord;
          With TableDefsRecordset do
               begin
               AddTextItem(NewRecIndex, 0, 'table');
               AddTextItem(NewRecIndex, 1, '0');
               AddTextItem(NewRecIndex, 2, 'sqlite_master');
               AddTextItem(NewRecIndex, 3, 'sqlite_master');
               AddTextItem(NewRecIndex, 4, '');
               AddTextItem(NewRecIndex, 5, '');
               end;

          for i := 0 to Pred(TableDefsRecordset.RecordsCount) do
              begin
              TableDef := TSqlitePassTableDef.Create(TableDefs);
              TableDef.FTableName := TableDefsRecordset.ItemsTextValue[i,2];
              { Sets the FieldDefs properties }
              GetDatabaseTableFieldDefs(TableDef);
              TableDef.FSql := TableDefsRecordset.ItemsTextValue[i,5];

              TableDefs.Add(TableDef);
              end;

          finally
          TableDefsRecordset.Free;
          end;
          end;
  end;

begin
TableDefs.ClearAndFreeItems;
GetTableDefs(nil);
{ Retrieves TableDefs info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count)
    do GetTableDefs(FDatabase.Databases[i]);
end;


procedure TSqlitePassTranslator.GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef);
var
i, NewRecIndex: integer;
Sql: String;
NewFieldDef: TSqlitePassTableFieldDef;
FieldDefsRecordset: TSqlitePassRecordset;

begin
TableDef.FieldDefs.Clear;
Sql := 'PRAGMA table_info ("' + TableDef.TableName +  '");';


FieldDefsRecordset := TSqlitePassRecordset.Create(FDatabase);
FieldDefsRecordset.Open(Sql);

{  We manually add the sqlite_master table which is not returned by engine }
NewRecIndex := FieldDefsRecordset.AddRecord;
With FieldDefsRecordset do
     begin
     AddTextItem(NewRecIndex, 1, 'rowid');
     AddTextItem(NewRecIndex, 2, 'INTEGER');
     AddTextItem(NewRecIndex, 3, '');
     AddTextItem(NewRecIndex, 4, '');
     AddTextItem(NewRecIndex, 5, '1');
     end;

for i := 0 to Pred(FieldDefsRecordset.RecordsCount) do
    begin
    NewFieldDef := TSqlitePassTableFieldDef.Create(TableDef.FieldDefs);
    With NewFieldDef do
         begin
         With FieldDefsRecordset do
              begin
              FFieldName := ItemsTextValue[i,1];
              FNativeDataType := ItemsTextValue[i,2];
              FNotNull := ItemsTextValue[i,3] <> '0';
              FDefaultValue := ItemsTextValue[i,4];
              FPrimaryKey := ItemsTextValue[i,5] = '1';
              FieldTypesTranslationRules.SetFieldDataType(NewFieldDef);

              { If a field is an Integer and Primary Key then it is considered
                as an Int64 AutoInc Field }
              If (FDataType = ftInteger) and FPrimaryKey then
                 begin;
                 FAutoInc  := True;
                 FDataType := ftLargeInt;
                 end;

              { Now we have to set the field size and precision }
              Case FDataType of
                   ftString : GetFieldDefSize(NewFieldDef);
                   ftBCD    : FPrecision := 2; // TODO Rework
                   end;
              end;
         end;
    TableDef.FieldDefs.Add(NewFieldDef);
    end;
FieldDefsRecordset.Free;
end;

procedure TSqlitePassTranslator.GetDatabaseIndexDefs(IndexDefs: TSqlitePassDatabaseIndexDefs);
var
i: Integer;
Sql, SqlIndexDefs, SqlIndexList, SqlIndexColumns: String;
IndexDefsRecordset: TSqlitePassRecordset;

  Procedure GetIndexDefs(AttachedDb: TSqlitePassDatabaseAttached = nil);
  var
  i, j, SubStringBegin: Integer;
  TempDb: TSqlitePassDatabase;
  DirectionStr: String;
  Idx: TSqlitePassIndex;
  Col: TSqlitePassIndexColumn;
  begin
    if Assigned(AttachedDb)
     then begin
          TempDb := TSqlitePassDatabase(FDatabase.FDatabasesList[FDatabase.FDatabases.IndexOf(AttachedDb)]);
          IndexDefs.AddIndexes(TempDb.IndexDefs, AttachedDb);
          end
     else begin
          { We get info on IndexName, TableName and SQLCreateStmt }
          IndexDefsRecordset.Open(SqlIndexDefs);
          for i := 0 to Pred(IndexDefsRecordset.RecordsCount) do
              begin
              Idx := TSqlitePassDatabaseIndex.Create(FDatabase.IndexDefs);
              Idx.FIndexName := IndexDefsRecordset.ItemsTextValue[i,2];
              Idx.FTableName := IndexDefsRecordset.ItemsTextValue[i,3];
              Idx.FSql := IndexDefsRecordset.ItemsTextValue[i,5];
              IndexDefs.Add(Idx);
              end;

          { For each table, we check if the index is unique or not }
          for i := 0 to Pred(FDatabase.TableDefs.Count) do
              begin
              Sql := Format(SqlIndexList, [FDatabase.TableDefs[i].TableName]);
              IndexDefsRecordset.Open(Sql);
              for j := 0 to Pred(IndexDefsRecordset.RecordsCount) do
                  begin
                  Idx := IndexDefs.FindIndex(IndexDefsRecordset.ItemsTextValue[j,1]);
                  if Assigned(Idx)
                     then Idx.Unique := (IndexDefsRecordset.ItemsTextValue[j,2]= '1');
                  end;
              end;

          { For each index, we get info about fields used and sort order }
          for i := 0 to Pred(IndexDefs.Count) do
              begin
              Sql := Format(SqlIndexColumns, [IndexDefs[i].IndexName]);
              IndexDefsRecordset.Open(Sql);
              for j := 0 to Pred(IndexDefsRecordset.RecordsCount) do
                begin
                Col := TSqlitePassIndexColumn.Create;
                Col.Position := StrToInt(IndexDefsRecordset.ItemsTextValue[j,1]);
                Col.ColumnName := IndexDefsRecordset.ItemsTextValue[j,2];
                SubStringBegin := Pos('"'+Col.ColumnName+'"', IndexDefs[i].Sql);
                If SubStringBegin <> -1 then
                   begin
                   DirectionStr := Copy(IndexDefs[i].Sql, SubStringBegin + 2 + length(Col.ColumnName), 6);
                   If Pos('ASC',DirectionStr) <> 0
                      then Col.Direction := cidAscending
                      else if Pos('DESC',DirectionStr) <> 0
                              then Col.Direction := cidDescending
                              else Col.Direction := cidUnknown;
                   end;
                IndexDefs[i].Columns.Add(Col);
                end;
               end;
          IndexDefsRecordset.Close;
          end; {Else begin}
  end; {GetIndexDefs}

begin
SqlIndexDefs := 'SELECT Rowid, * FROM sqlite_master WHERE type = "index" ' +
                'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "index" ' +
                'ORDER BY type, tbl_name;';
SqlIndexList := 'PRAGMA index_list ("%s");';
SqlIndexColumns := 'PRAGMA index_info ("%s");';

IndexDefs.ClearAndFreeItems;
IndexDefsRecordset:= TSqlitePassRecordset.Create(FDatabase);
try
  GetIndexDefs(nil);
  { Retrieves IndexDefs info for attached databases }
  For i := 0 to Pred(FDatabase.Databases.Count)
      do GetIndexDefs(FDatabase.Databases[i]);
finally
  IndexDefsRecordset.Free;
end;
end;

function TSqlitePassTranslator.IsSystemTable(TableName: String): Boolean;
begin
Result :=    (CompareText(TableName, SqlitePassDbSettingsTable) = 0)
          or (CompareText(TableName, SqlitePassSQLStmtDefsTable) = 0)
          or (CompareText(TableName, 'sqlite_master') = 0);
end;

procedure TSqlitePassTranslator.ConvertNativeFieldTypeToPascalFieldType(FieldDef: TSqlitePassTableFieldDef);
begin
{ Implementation should be done in TSqlitePassTranslator descendants }
end;

procedure TSqlitePassTranslator.GetFieldDefSize(FieldDef: TSqlitePassTableFieldDef);
var
StartPos, EndPos: Word;
begin
FieldDef.FSize := 0;
StartPos:= Pos('(', FieldDef.FNativeDataType);
if StartPos > 0 then
   begin
   EndPos:=pos(')', FieldDef.FNativeDataType);
   if EndPos > StartPos
      then FieldDef.FSize :=
           Trunc(StrToFloat(Copy(FieldDef.FNativeDataType, StartPos+1, EndPos - StartPos - 1)));
   end;
end;


procedure TSqlitePassTranslator.GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs);
begin
QueryDefs.ClearAndFreeItems;
{ QueryDefs are not stored by sqlite engine, so there is nothing to do
  Implementation must be done in TSqlitePassTranslator descendants }
end;

procedure TSqlitePassTranslator.GetDatabaseViews(Views: TSqlitePassViews);
const
Sql = 'SELECT Rowid, * FROM sqlite_master WHERE type = "view" ' +
      'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "view" ' +
      'ORDER BY type, tbl_name;';
var
i: Integer;

  Procedure GetViews(AttachedDb: TSqlitePassDatabaseAttached = nil);
  var
  CreateStmtStart, i: Integer;
  TempDb: TSqlitePassDatabase;
  View: TSqlitePassView;
  ViewsRecordset: TSqlitePassRecordset;

  begin

  if Assigned(AttachedDb)
     then begin
          TempDb := TSqlitePassDatabase(FDatabase.FDatabasesList[FDatabase.FDatabases.IndexOf(AttachedDb)]);
          Views.AddViews(TempDb.Views, AttachedDb);
          end
     else begin
          TempDb := FDatabase;
          ViewsRecordset := TSqlitePassRecordset.Create(TempDb);

          try
          ViewsRecordset.Open(Sql);

          for i := 0 to Pred(ViewsRecordset.RecordsCount) do
              begin
              View := TSqlitePassView.Create(Views);
              View.FViewName := ViewsRecordset.ItemsTextValue[i,3];
              View.FSqlCreateStmt := ViewsRecordset.ItemsTextValue[i,5];
              CreateStmtStart := Pos(kwSelect + ' ', Uppercase(View.FSqlCreateStmt));

              if CreateStmtStart <> 0
                 then View.FSqlSelectStmt := Copy(View.FSqlCreateStmt, CreateStmtStart, Length(View.FSqlCreateStmt));

              Views.Add(View);
              end;

          finally
          ViewsRecordset.Free;
          end;
          end;
    end;


begin
Views.ClearAndFreeItems;
GetViews(nil);
{ Retrieves Views info for attached databases }
for i := 0 to Pred(FDatabase.Databases.Count)
    do GetViews(FDatabase.Databases[i]);
end;


procedure TSqlitePassTranslator.GetDatabaseTriggers(Triggers: TSqlitePassTriggers);
var
i: Integer;
Sql: String;
TriggersRecordset: TSqlitePassRecordset;

  Procedure GetTriggers(DatabaseConnection: Pointer);
  var
  i: integer;
  Trigger: TSqlitePassTrigger;

  begin
  TriggersRecordset.Open(Sql);

  for i := 0 to Pred(TriggersRecordset.RecordsCount)
   do begin
      Trigger := TSqlitePassTrigger.Create(Triggers);
      Trigger.FTriggerName := TriggersRecordset.ItemsTextValue[i,2];
      Trigger.FSQL := TriggersRecordset.ItemsTextValue[i,3];
      //Trigger.FAttached := False;  // TODO could be True too...
      Triggers.Add(Trigger);
      end;

  TriggersRecordset.Close;
  end;

begin
Sql := 'SELECT Rowid, * FROM sqlite_master WHERE type = "trigger" ' +
       'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "trigger" ' +
       'ORDER BY type, tbl_name;';
Triggers.ClearAndFreeItems;
TriggersRecordset := TSqlitePassRecordset.Create(FDatabase);
GetTriggers(FDatabase.Engine.Connection);
{ Retrieves Triggers info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count) do
    begin
//    FDatabase.Databases[i].FEngine.OpenDatabase(FDatabase.Databases[i].Database);
//    GetTriggers(FDatabase.Databases[i].FEngine.Connection);
//    FDatabase.Databases[i].FEngine.CloseDatabase;
    end;
TriggersRecordset.Free;
end;



{ TSqlitePassTranslator_Kexi }

function TSqlitePassTranslator_Kexi.GetNativeFieldTypeName(FieldType: Word): String;
begin
Case FieldType of
     // TODO - Result := ftBytes doesn't return correct data - Changed to ftSmallInt;
     1: Result := 'byte';
     2: Result := 'shortinteger';
     3: Result := 'integer';
     // TODO - LargeInt doesn't seem to be supported by kexi binding (Check autoInc field) 4: Result := ftLargeint;
     4: Result := 'biginteger';
     5: Result := 'boolean';
     6: Result := 'date';
     7: Result := 'datetime';
     8: Result := 'time';
     9: Result := 'float';
     10: Result := 'float';
     11: Result := 'text';
     12: Result := 'clob';
     13: Result := 'blob';// ftGraphic;
     else Result := 'unknown';
     end;
end;

procedure TSqlitePassTranslator_Kexi.ConvertNativeFieldTypeToPascalFieldType(FieldDef: TSqlitePassTableFieldDef);
begin
{ Translated from Kexi project : field.h }
Case FieldDef.FNativeDataTypeCode of
     { Byte = 1 }
     // TODO - Result := ftBytes doesn't return correct data - Changed to ftSmallInt;
     1: FieldDef.DataType := ftSmallInt;
     { ShortInteger = 2 }
     2: FieldDef.DataType := ftSmallInt;
     { Integer = 3 }
     3: FieldDef.DataType := ftInteger;
     { BigInteger = 4 }
     // TODO - LargeInt doesn't seem to be supported by kexi binding (Check autoInc field) 4: FieldDef.DataType := ftLargeint;
     4: FieldDef.DataType := ftLargeInt;
     { Boolean = 5 }
     5: FieldDef.DataType := ftBoolean;
     { Date = 6 }
     6: FieldDef.DataType := ftDate;
     { DateTime = 7 }
     7: FieldDef.DataType := ftDateTime;
     { Time = 8 }
     8: FieldDef.DataType := ftTime;
     { Float = 9 }
     9: FieldDef.DataType := ftFloat;
     { Double = 10 }
     10: FieldDef.DataType := ftFloat;
     { Text = 11 }
     11: FieldDef.DataType := ftString;
     { LongText = 12 }
     12: FieldDef.DataType := ftMemo;
     { BLOB = 13 }
     13: FieldDef.DataType := ftBlob; { Should be set to ftGraphic ? }
     else FieldDef.DataType := ftunknown;
     end;
end;

procedure TSqlitePassTranslator_Kexi.GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef);
const
SqlConst = 'SELECT t1.o_id,t1.o_name,t1.o_type,t2.f_type,t2.f_name,t2.f_length,'
    + 't2.f_precision,t2.f_constraints,t2.f_options,t2.f_default,t2.f_order,'
    + 't2.f_caption,t2.f_help '
    + 'FROM kexi__objects t1,kexi__fields t2 '
    + 'WHERE t1.o_id=t2.t_id AND t1.o_name = "%s";' ;
var
i: integer;
FieldConstraints: TSqlitePassBitArray;
aFieldDef: TSqlitePassTableFieldDef;
Sql: String;
FieldDefsRecordset: TSqlitePassRecordset;

begin
TableDef.FieldDefs.Clear;
FieldDefsRecordset := TSqlitePassRecordset.Create(FDatabase);
Sql := Format(SqlConst,[TableDef.TableName]);

Try
  Try
  FieldDefsRecordset.Open(Sql);
  FieldConstraints := TSqlitePassBitArray.Create;
  FieldConstraints.Items := KexiFieldConstraints;
  // TODO - Handle System Tables Kexi__...

  for i := 0 to Pred(FieldDefsRecordset.RecordsCount) do
      begin
      aFieldDef := TSqlitePassTableFieldDef.Create(TableDef.FieldDefs);
      With aFieldDef do
           begin
           With FieldDefsRecordset do
                begin
                FFieldName := ItemsTextValue[i,4];
                FDisplayName :=  ItemsTextValue[i,11];
                FNativeDataTypeCode := StrToInt(ItemsTextValue[i,3]);
                FNativeDataType := GetNativeFieldTypeName(FNativeDataTypeCode);
                FieldTypesTranslationRules.SetFieldDataType(aFieldDef);
                FTableFieldNo := StrToInt(ItemsTextValue[i,10]);
                FDefaultValue := ItemsTextValue[i,9];
                FHint := ItemsTextValue[i,12];

                { Now we have to set the field size and precision }
                Case AFieldDef.DataType of
                     ftString : FSize := StrToInt(ItemsTextValue[i,5]);
                     ftBCD    : FPrecision := StrToInt(ItemsTextValue[i,6]);
                     end;

                { Now we have to set the field constraints }
                FieldConstraints.Value := StrToInt(ItemsTextValue[i,7]);
                end; {With FieldDefsRecordset[i] }
            FNoConstraints := FieldConstraints.IsTrue('NoConstraints');
            FAutoInc := FieldConstraints.IsTrue('AutoInc');
            FUnique := FieldConstraints.IsTrue('Unique');
            FPrimaryKey := FieldConstraints.IsTrue('PrimaryKey');
            FForeignKey := FieldConstraints.IsTrue('ForeignKey');
            FNotNull := FieldConstraints.IsTrue('NotNull');
            FNotEmpty := FieldConstraints.IsTrue('NotEmpty');
            FIndexed := FieldConstraints.IsTrue('Indexed');
            end; { With aFieldDef }
      TableDef.FieldDefs.Add(aFieldDef);
      end; { for }

  { If we don't find any fieldefs (system tables fielddefs are not returned by kexi)
    we try the classic method... }
  if TableDef.FieldDefs.Count = 0
     then inherited GetDatabaseTableFieldDefs(TableDef);

  Except
  FDatabase.DatabaseError.RaiseExceptionFmt(Msg4000,['kexi__objects']);
  end;

Finally
FieldConstraints.Free;
FieldDefsRecordset.Close;
FieldDefsRecordset.Free;
end;
end;

procedure TSqlitePassTranslator_Kexi.GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs);
var
i: Integer;
Sql: String;
QueryDefsRecordset: TSqlitePassRecordset;

  Procedure GetQueryDefs(AliasName: String; DatabaseConnection: Pointer);
  var
  i: integer;
  QueryDef: TSqlitePassQueryDef;
  begin
  QueryDefsRecordset.Open(Sql);
  for i := 0 to Pred(QueryDefsRecordset.RecordsCount)
   do begin
      QueryDef := TSqlitePassQueryDef.Create(QueryDefs);
      QueryDef.FQueryName := QueryDefsRecordset.ItemsTextValue[i,3];
//      if AliasName <> ''
//         then QueryDef.FQueryFullName := AliasName + '.' + QueryDef.FQueryName
//         else QueryDef.FQueryFullName := QueryDef.FQueryName;
      QueryDef.FSql := QueryDefsRecordset.ItemsTextValue[i,4];
      QueryDefs.Add(QueryDef);
      end;
  QueryDefsRecordset.Close;
  end;

begin
Sql := 'SELECT K1.Rowid, K2.rowid, K1.o_type,K1.o_name,K2.o_data '+
       'FROM kexi__objects K1, kexi__objectdata K2 '+
       'WHERE (K1.o_id=K2.o_id) AND (K1.o_type=2) AND (K2.o_sub_id="sql");';
QueryDefs.ClearAndFreeItems;

QueryDefsRecordset := TSqlitePassRecordset.Create(FDatabase);
GetQueryDefs('', FDatabase.Engine.Connection);

{ Retrieves TableDefs info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count) do
    begin
//    FDatabase.Databases[i].FEngine.OpenDatabase(FDatabase.Databases[i].Database);
//    GetQueryDefs(FDatabase.Databases[i].AliasName, FDatabase.Databases[i].FEngine.Connection);
//    FDatabase.Databases[i].FEngine.CloseDatabase;
    end;

QueryDefsRecordset.Free;
end;

function TSqlitePassTranslator_Kexi.IsSystemTable(
  TableName: String): Boolean;
var
i: Integer;
begin
for i := 0 to High(KexiSystemTables) do
    begin
    Result :=  (CompareText(TableName, KexiSystemTables[i]) = 0);
    if Result then break;
    end;
If Not Result
   then Result := Inherited IsSystemTable(TableName);//((CompareTableName = SqlitePassSystemTable)
end;


{ SQLiteExpert - TSqlitePassTranslator_sqliteExpert }
{ Nothing to override...}

{ SQLite Administrator - TSqlitePassTranslator_sqliteAdmin }
{ Nothing to override...}



{ TSqlitePassDatabaseDataTypeOptions }

constructor TSqlitePassDatabaseDataTypeOptions.Create(
  Database: TSqlitePassDatabase);
begin
  Inherited Create;
  FDatabase := Database;
  FDecimalSeparator := SysUtils.DecimalSeparator;
  DateFormat := 'YYYY-MM-DD';
  TimeFormat := 'hh:mm:ss';
  DateTimeFormat := 'YYYY-MM-DD hh:mm:ss.zzz';
  FLoadOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
  FSaveOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
  FFieldTypesTranslationRules := TSqlitePassFieldTypesTranslationRules.Create(Self);;
  FCustomFieldDefs := TSqlitePassCustomFieldDefs.Create;
  FBooleanTrueValues := TStringList.Create;
  FBooleanFalseValues := TStringList.Create;
end;

destructor TSqlitePassDatabaseDataTypeOptions.Destroy;
begin
  FBooleanTrueValues.Free;
  FBooleanFalseValues.Free;
  FCustomFieldDefs.Free;
  FFieldTypesTranslationRules.Free;
  Inherited Destroy;
end;

function TSqlitePassDatabaseDataTypeOptions.GetOwner: TPersistent;
begin
  Result := FDatabase;
end;

procedure TSqlitePassDatabaseDataTypeOptions.LoadFromDatabase(_LoadOptions: TsqlitePassDataTypeStorageOptions = []);
Const
Sql = 'SELECT * FROM "' + SqlitePassDbSettingsTable +'";';
var
i: Integer;
DbSettingsRecordset: TSqlitePassRecordset;
StorageInfoStrings, MainPropertiesStrings, TranslationRulesStrings, CustomFieldDefsStrings: TStringList;
TranslationRule: TSqlitePassFieldTypeTranslationRule;
CustomFieldDef: TSqlitePassCustomFieldDef;
LoadingError: Boolean;
begin
If Not FDatabase.Connected then Exit;

LoadingError := False;

Try {1}
  If _LoadOptions = []
     then _LoadOptions := FLoadOptions;

  StorageInfoStrings      := TStringList.Create;
  MainPropertiesStrings   := TStringList.Create;
  TranslationRulesStrings := TStringList.Create;
  CustomFieldDefsStrings  := TStringList.Create;

  { If the table doesn't exist, jump to the finally part }
  If (FDatabase.TableDefs.FindTable(SqlitePassDbSettingsTable) = nil)
     then Exit;

  Try {2}

     DbSettingsRecordset:= TSqlitePassRecordset.Create(FDatabase);
     Try {3}
     { Loads Options from database }
     DbSettingsRecordset.Open(Sql);
        If (DbSettingsRecordset.RecordsCount > 0) then
        begin
        StorageInfoStrings.SetText(PChar(DbSettingsRecordset.ItemsTextValue[0,0]));
        MainPropertiesStrings.SetText(PChar(DbSettingsRecordset.ItemsTextValue[0,1]));
        TranslationRulesStrings.SetText(PChar(DbSettingsRecordset.ItemsTextValue[0,2]));
        CustomFieldDefsStrings.SetText(PChar(DbSettingsRecordset.ItemsTextValue[0,3]));
        end
        else LoadingError := True;
     finally
       DbSettingsRecordset.Close;
       DbSettingsRecordset.Free;
     end; { Try 3}

   { Table SQLitePass__DbSettings doesn't exist...}
   if LoadingError then exit;

   { Storage Version }
   if StorageInfoStrings.Count = 2
      then FStorageVersion := StorageInfoStrings[1];

   if FStorageVersion <> SqlitePassStorageVersion
      then Abort;

   { Main Properties }
   if soProperties in _LoadOptions then
      begin
      BooleanFormat := MainPropertiesStrings[1];
      BooleanStorage := TSqlitePassBooleanStorage(StrToInt(MainPropertiesStrings[3]));
      DateFormat := MainPropertiesStrings[5];
      DateStorage := TSqlitePassDateStorage(StrToInt(MainPropertiesStrings[7]));
      DateTimeFormat := MainPropertiesStrings[9];
      DateTimeStorage := TSqlitePassDateTimeStorage(StrToInt(MainPropertiesStrings[11]));
      DecimalSeparator := MainPropertiesStrings[13][1];
      DefaultFieldType := StringToFieldType(MainPropertiesStrings[15]);
      DetectionMode := TSqlitePassDataTypeDetectionMode(StrToInt(MainPropertiesStrings[17]));
      TimeFormat := MainPropertiesStrings[19];
      TimeStorage := TSqlitePassTimeStorage(StrToInt(MainPropertiesStrings[21]));
       end;

   { TranslationRules }
   if soTranslationRules in _LoadOptions then
      begin
      TranslationRules.ClearAndFreeItems;
      i := 1;
      While i < Pred(TranslationRulesStrings.Count) do
            begin
            TranslationRule := TSqlitePassFieldTypeTranslationRule.Create;
            TranslationRule.DatatypeName := TranslationRulesStrings[i];
            Inc(i,2);
            TranslationRule.MappingMode :=  TSqlitePassDataTypeMappingMode(StrToInt(TranslationRulesStrings[i]));
            Inc(i,2);
            TranslationRule.FieldType := StringToFieldType(TranslationRulesStrings[i]);
            TranslationRules.Add(TranslationRule);
            Inc(i,2);
            end;
       end;

   { CustomFieldDefs }
   if soCustomFieldDefs in _LoadOptions then
      begin
      CustomFieldDefs.ClearAndFreeItems;
      i := 1;
      While i < Pred(CustomFieldDefsStrings.Count) do
            begin
            CustomFieldDef := TSqlitePassCustomFieldDef.Create;
            CustomFieldDef.TableName := CustomFieldDefsStrings[i];
            Inc(i,2);
            CustomFieldDef.FieldName := CustomFieldDefsStrings[i];
            Inc(i,2);
            CustomFieldDef.FieldType := StringToFieldType(CustomFieldDefsStrings[i]);
            Inc(i,2);
            CustomFieldDef.FieldSize := StrToInt(CustomFieldDefsStrings[i]);
            Inc(i,2);
            CustomFieldDef.FieldPrecision := StrToInt(CustomFieldDefsStrings[i]);
            CustomFieldDefs.Add(CustomFieldDef);
            Inc(i,2);
            end;
      end;
   except
      if FDatabase.DatabaseError.ShowError(Msg1044, mtWarning, [mbYes, mbNo], 0) = mrYes
         then FDatabase.Engine.ExecSQL('Drop table "SQLitePass__DbSettings";');
   end; {Try 2 }

 finally
 StorageInfoStrings.Free;
 MainPropertiesStrings.Free;
 TranslationRulesStrings.Free;
 CustomFieldDefsStrings.Free;
 end; {Try 1}
end;


procedure TSqlitePassDatabaseDataTypeOptions.SaveToDatabase(_SaveOptions: TsqlitePassDataTypeStorageOptions = []);
Const
Sql = 'CREATE TABLE IF NOT EXISTS "' + SqlitePassDbSettingsTable +'" ("StorageVersion" CBLOB, "MainProperties" CBLOB, "TranslationRules" CBLOB, "CustomFieldDefs" CBLOB);';
var
i: Integer;
WasConnected, WasReadOnly: Boolean;
DbSettingsDataset: TSqlitePassDataset;
DbSettingTableDef: TSqlitePassTableDef;
Lines: TStringList;
begin
If Assigned(FDatabase) and Assigned(FDatabase.FEngine) then
   begin
   try
     DbSettingsDataset := nil;
     if _SaveOptions = []
        then _SaveOptions := FSaveOPtions;

     Lines := TStringList.Create;
     WasConnected := FDatabase.Connected;
     WasReadOnly  := FDatabase.ReadOnly;
     FDatabase.ReadOnly := False;
     FDatabase.Open;

     { Creates the table if it doesn't already exist }
     DbSettingTableDef := FDatabase.FTableDefs.FindTable(SqlitePassDbSettingsTable);
     if DbSettingTableDef = nil then
        begin
        FDatabase.FEngine.ExecSQL(Sql);
        FDatabase.Translator.GetDatabaseTableDefs(FDatabase.TableDefs);
        DbSettingTableDef := FDatabase.FTableDefs.FindTable(SqlitePassDbSettingsTable)
        end;

     { Enforces fields fieldTypes to ftMemo whatever Databasetype is used }
     For i := 0 to Pred(DbSettingTableDef.FieldDefs.Count)
         do DbSettingTableDef.FieldDefs[i].DataType := ftMemo;

     { Saves Options to database }
     DbSettingsDataset:= TSqlitePassDataset.Create(nil);
     DbSettingsDataset.Database := FDatabase;

     With DbSettingsDataset do
          begin
          DatasetName := SqlitePassDbSettingsTable;
          Open;
          First;
          Edit;
          end;

    { Save Storage Version }
     With Lines do
          begin
          Add('StorageVersion');Add(SqlitePassStorageVersion);
          end;
     DbSettingsDataset.Fields[0].Assign(Lines);

     { Main Properties }
     if soProperties in _SaveOptions then
        begin
        With Lines do
             begin
             Clear;
             Add('BooleanExtension'); Add(BooleanFormat);
             Add('BooleanStorage'); Add(IntToStr(Ord(BooleanStorage)));
             Add('DateFormat'); Add(DateFormat);
             Add('DateStorage'); Add(IntToStr(Ord(DateStorage)));
             Add('DateTimeFormat'); Add(DateTimeFormat);
             Add('DateTimeStorage'); Add(IntToStr(Ord(DateTimeStorage)));
             Add('DecimalSeparator'); Add(DecimalSeparator);
             Add('DefaultFieldType'); Add(FieldTypeToString(DefaultFieldType));
             Add('DetectionMode'); Add(IntToStr(Ord(DetectionMode)));
             Add('TimeFormat'); Add(TimeFormat);
             Add('TimeStorage'); Add(IntToStr(Ord(TimeStorage)));
             DbSettingsDataset.Fields[1].Assign(Lines);
             end;
        end;

     { TranslationRules }
     if soTranslationRules in _SaveOptions then
        begin
          With Lines do
          begin
          Clear;
          for i := 0 to Pred(TranslationRules.Count) do
              begin
              Add('DatatypeName');
              Add(TranslationRules[i].DatatypeName);
              Add('MappingMode');
              Add(IntToStr(Ord(TranslationRules[i].MappingMode)));
              Add('FieldType');
              Add(FieldTypeToString(TranslationRules[i].FieldType));
              end;
          end;
        DbSettingsDataset.Fields[2].Assign(Lines);
        end;

     { Custom FieldDefs }
      if soCustomFieldDefs in _SaveOptions then
        begin
          With Lines do
          begin
          Clear;
          for i := 0 to Pred(CustomFieldDefs.Count) do
              begin
              Add('TableName'); Add(CustomFieldDefs[i].TableName);
              Add('FieldName'); Add(CustomFieldDefs[i].FieldName);
              Add('FieldType'); Add(FieldTypeToString(CustomFieldDefs[i].FieldType));
              Add('FieldSize'); Add(IntToStr(CustomFieldDefs[i].FieldSize));
              Add('FieldPrecision'); Add(IntToStr(CustomFieldDefs[i].FieldPrecision));
              end;
          end;
        DbSettingsDataset.Fields[3].Assign(Lines);
        end;

   finally
     if Assigned(DbSettingsDataset) then
        begin
        DbSettingsDataset.Post;
        DbSettingsDataset.Free;
        end;
     Lines.Free;
     FDatabase.ReadOnly  := WasReadOnly;
     FDatabase.Connected := WasConnected;
   end;
   end;
end;

procedure TSqlitePassDatabaseDataTypeOptions.DefineProperties(Filer: TFiler);
begin
{ Property are prefixed with 'p' to have a different name of the original
  property otherwise, DefineProperty is not called by the filer }
  Filer.DefineProperty('pCustomFieldDefs', ReadCustomFieldDefs, WriteCustomFieldDefs, True);
  Filer.DefineProperty('pTranslationsRules', ReadTranslationRules, WriteTranslationRules, True);
end;

procedure TSqlitePassDatabaseDataTypeOptions.ReadTranslationRules(Reader: TReader);
var
TranslationRule: TSqlitePassFieldTypeTranslationRule;
begin
FFieldTypesTranslationRules.ClearAndFreeItems;
Reader.ReadListBegin;
while not(Reader.EndofList) do
      begin
      TranslationRule := TSqlitePassFieldTypeTranslationRule.Create;
      With TranslationRule do
           begin
           DatatypeName := Reader.ReadString;
           MappingMode := TSqlitePassDataTypeMappingMode(Reader.ReadInteger);
           FieldType := StringToFieldType(Reader.ReadString);
           end;
      FFieldTypesTranslationRules.Add(TranslationRule);
      end;
Reader.ReadListEnd;
end;

procedure TSqlitePassDatabaseDataTypeOptions.WriteTranslationRules(Writer: TWriter);
var
  i: Integer;
begin
  With Writer do
       begin
       WriteListBegin;
       for i := 0 to Pred(FFieldTypesTranslationRules.Count) do
          begin
          WriteString(FFieldTypesTranslationRules[i].DatatypeName);
          WriteInteger(Ord(FFieldTypesTranslationRules[i].MappingMode));
          WriteString(FieldTypeToString(FFieldTypesTranslationRules[i].FieldType));
          end;
       WriteListEnd;
       end;
end;

procedure TSqlitePassDatabaseDataTypeOptions.ReadCustomFieldDefs(Reader: TReader);
var
CustomFieldDef: TSqlitePassCustomFieldDef;
begin
FCustomFieldDefs.ClearAndFreeItems;
Reader.ReadListBegin;
while not(Reader.EndofList) do
      begin
      CustomFieldDef := TSqlitePassCustomFieldDef.Create;
      With CustomFieldDef do
           begin
           TableName := Reader.ReadString;
           FieldName := Reader.ReadString;
           FieldType := StringToFieldType(Reader.ReadString);
           FieldSize := Reader.ReadInteger;
           FieldPrecision := Reader.ReadInteger;
           end;
      FCustomFieldDefs.Add(CustomFieldDef);
      end;
Reader.ReadListEnd;
end;

procedure TSqlitePassDatabaseDataTypeOptions.WriteCustomFieldDefs(Writer: TWriter);
var
  i: Integer;
begin
  With Writer do
       begin
       WriteListBegin;
       for i := 0 to Pred(FCustomFieldDefs.Count) do
          begin
          WriteString(FCustomFieldDefs[i].TableName);
          WriteString(FCustomFieldDefs[i].FieldName);
          WriteString(FieldTypeToString(FCustomFieldDefs[i].FieldType));
          WriteInteger(FCustomFieldDefs[i].FieldSize);
          WriteInteger(FCustomFieldDefs[i].FieldPrecision);
          end;
       WriteListEnd;
       end;
end;

procedure TSqlitePassDatabaseDataTypeOptions.SetDefaultValues(DatabaseType: TSqlitePassDatabaseType; SetTranslationRules: Boolean=false);
begin

{ Generic Default Values }
FDetectionMode := dmTypeName; { dmTypeName, dmDbSpecific, dmCustom, dmForceStr }
FDateStorage   := asInteger;
FDateTimeStorage := dtsText;
FTimeStorage := asInteger;
FBooleanStorage := asInteger;
FDecimalSeparator := '.';
FDefaultFieldType := ftUnknown;
DateFormat := 'YYYY-MM-DD';
TimeFormat := 'hh:mm:ss';
DateTimeFormat := 'YYYY-MM-DD hh:mm:ss';
BooleanFormat := 'True,Yes,On,1,-False,No,Off,0';
if FLoadOptions = []
   then FLoadOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
if FSaveOptions = []
   then FSaveOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
FDatabase.FOptions.FQuoteStyle := qsDoubleQuote;

{ Overrides default values depending on DatabaseType }
Case DatabaseType of
  dbtSqlitePass:
           begin
           DateTimeStorage := dtsDateTime;
           DateTimeFormat := 'YYYY-MM-DD hh:mm:ss.zzz';
           BooleanFormat := '1-0'
           end;

  dbtKexi: begin
           FDetectionMode := dmDbSpecific;
           DateTimeFormat := 'YYYY-MM-DDThh:mm:ss';
           BooleanFormat := '1-0'
           end;

  dbtSqliteAdmin:
           begin
           FDatabase.FOptions.FQuoteStyle := qsBracket;
           BooleanFormat := '1-0'
           end;

  dbtSqliteExpert:
           begin
           FDatabase.FOptions.FQuoteStyle := qsBracket;
           FDateStorage   := asText;
           FTimeStorage := asText;
           BooleanFormat := '1-0'
           end;

  dbtSqlite4FPC:
           begin
           FDefaultFieldType := ftString;
           end;
  end;
if SetTranslationRules
   then TranslationRules.SetDefaultTranslationRules(DatabaseType);
end;

procedure TSqlitePassDatabaseDataTypeOptions.ApplyCustomFieldDefs;
var
  i: Integer;
  FieldDef: TSqlitePassTableFieldDef;
begin
for i:= 0 to Pred(FCustomFieldDefs.Count) do
    begin
    FieldDef := FDatabase.TableDefs.FindFieldDef(FCustomFieldDefs[i].TableName,FCustomFieldDefs[i].FieldName);
    if Assigned(FieldDef)
       then FieldDef.DataType := FCustomFieldDefs[i].FieldType;
    end;
end;

procedure TSqlitePassDatabaseDataTypeOptions.FindSubString
(Chr: Char; LowerStr: String; Var StartPos, EndPos: Word);
var
StrLength: Integer;
begin
StartPos := Pos(Chr,LowerStr);
StrLength := Length(LowerStr);
if StartPos > 0 then
   begin
   EndPos := StartPos;
   While (LowerStr[EndPos] = Chr) and (EndPos <= StrLength) Do Inc(EndPos);
   end;
Dec(EndPos,StartPos);
end;

procedure TSqlitePassDatabaseDataTypeOptions.SetFDateFormat(const Value: String);
var
LowerStr: String;
begin
{Year-Month-Day
FDateFormatPattern : Array[0..2, 0..1] of Word;}
LowerStr := Lowercase(Value);
FindSubString('y', LowerStr, FDateFormatPattern[0,0], FDateFormatPattern[0,1]);
FindSubString('m', LowerStr, FDateFormatPattern[1,0], FDateFormatPattern[1,1]);
FindSubString('d', LowerStr, FDateFormatPattern[2,0], FDateFormatPattern[2,1]);
FDateFormat := Value;
end;

procedure TSqlitePassDatabaseDataTypeOptions.SetFTimeFormat(const Value: String);
var
LowerStr: String;
begin
{Hour-Minute-Sec-Msec
FTimeFormatPattern : Array[0..3, 0..1] of Word;}
LowerStr := Lowercase(Value);
FindSubString('h', LowerStr, FTimeFormatPattern[0,0], FTimeFormatPattern[0,1]);
FindSubString('m', LowerStr, FTimeFormatPattern[1,0], FTimeFormatPattern[1,1]);
FindSubString('s', LowerStr, FTimeFormatPattern[2,0], FTimeFormatPattern[2,1]);
FindSubString('z', LowerStr, FTimeFormatPattern[3,0], FTimeFormatPattern[3,1]);
FTimeFormat := Value;
end;

procedure TSqlitePassDatabaseDataTypeOptions.SetFDateTimeFormat(const Value: String);
begin
{Year-Month-Day-Hour-Minute-Sec-Msec
FDateTimeFormatPattern : Array[0..6, 0..1] of Word;

DateTimeFormat is Case Sensitive
}
FindSubString('Y', Value, FDateTimeFormatPattern[0,0], FDateTimeFormatPattern[0,1]);
FindSubString('M', Value, FDateTimeFormatPattern[1,0], FDateTimeFormatPattern[1,1]);
FindSubString('D', Value, FDateTimeFormatPattern[2,0], FDateTimeFormatPattern[2,1]);
FindSubString('h', Value, FDateTimeFormatPattern[3,0], FDateTimeFormatPattern[3,1]);
FindSubString('m', Value, FDateTimeFormatPattern[4,0], FDateTimeFormatPattern[4,1]);
FindSubString('s', Value, FDateTimeFormatPattern[5,0], FDateTimeFormatPattern[5,1]);
FindSubString('z', Value, FDateTimeFormatPattern[6,0], FDateTimeFormatPattern[6,1]);
FDateTimeFormat := Value;
end;

procedure TSqlitePassDatabaseDataTypeOptions.SetFBooleanFormat(const Value: String);
var
TrueStr, FalseStr: String;
P: Integer;
begin
P := Pos('-',Value);
TrueStr  := Copy(Value,1, Pred(P));
if TrueStr = ''
   then TrueStr := '1';
FalseStr := Copy(Value,Succ(P), MaxInt);
if FalseStr = ''
   then FalseStr := '0';
FBooleanTrueValues.CommaText := TrueStr;
FBooleanFalseValues.CommaText := FalseStr;
FBooleanFormat := Value;
end;

{ TSqlitePassDatabaseError }

constructor TSqlitePassDatabaseError.Create(Database: TSqlitePassDatabase);
begin
 Inherited Create;
 FDatabase := Database;
 FErrorStringList := TStringList.Create;
 FErrorCount := 0;
 FLastErrorCode := -1;
 FLastErrorMsg := Msg1002;
 FLastErrorTime := Now();
end;

destructor TSqlitePassDatabaseError.Destroy;
begin
 FDatabase := nil;
 FErrorStringList.Free;
 Inherited Destroy;
end;

procedure TSqlitePassDatabaseError.GetLastError(var ErrorCode: Integer; var Msg: String;
 var Time: TDateTime);
begin
 ErrorCode := FLastErrorCode;
 Msg  := FLastErrorMsg;
 Time := FLastErrorTime;
end;

procedure TSqlitePassDatabaseError.LogError(Msg: string; const Args: array of const; const ErrorCode: Integer = -1; Component: TComponent = nil);
begin
 FLastErrorCode := ErrorCode;
 if High(Args) = -1
    then FLastErrorMsg := Msg
    else FLastErrorMsg := Format(Msg, Args);

 if Component is TSqlitePassDataset
    then Msg := 'Dataset : ' + TSqlitePassDataset(Component).DatasetName + ' : ' + Msg;

 FLastErrorTime := Now();
 
 if FDatabase.FOptions.LogErrors
    then with FErrorStringList do
         begin
         Inc(FErrorCount);
         Add('--- Error ' + IntToStr(FErrorCount) + ' - ' + DateTimeToStr(FLastErrorTime) + ' Begin ---');
         Add('    Message : ' + FLastErrorMsg);
         Add('    Error code : ' + IntToStr(FLastErrorCode));
         Add('--- Error ' + IntToStr(FErrorCount) + ' End ---');
         Add('');
         end;
{$IFDEF DEBUG}
 FDatabase.Debugger.Display(FLastErrorMsg);
{$ENDIF}
end;

procedure TSqlitePassDatabaseError.RaiseException(const Msg: String; const ErrorCode: Integer = -1;  Component: TComponent = nil; VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow);
begin
 Case VerboseLevel of
      vlLog        : LogError(Msg, [], ErrorCode, Component);
      vlShow       : DatabaseError(Msg, Component);
      vlLogAndShow : begin
                     LogError(Msg, [], ErrorCode, Component);
                     DatabaseError(Msg, Component);
                     end;
      else Abort;
      end;
end;

procedure TSqlitePassDatabaseError.RaiseExceptionFmt(const Msg: string; const Args: array of const; const ErrorCode: Integer = -1; Component: TComponent = nil; VerboseLevel: TSqlitePassVerboseLevel = vlLogAndShow);
begin
 If High(Args) = -1
    then RaiseException(Msg, ErrorCode, Component, VerboseLevel)
    else RaiseException(Format(Msg, Args), ErrorCode, Component, VerboseLevel);
end;

procedure TSqlitePassDatabaseError.SaveToFile(FileName: String);
begin
 FErrorStringList.SaveToFile(FileName);
end;

function TSqlitePassDatabaseError.ShowError(const Msg: string; DlgType: TMsgDlgType = mtWarning;
 DlgButtons: TMsgDlgButtons = [mbOk]; HelpCtx: Longint = 0): Word;
begin
 LogError(Msg, []);
 Result := MessageDlg(Msg, DlgType, DlgButtons, HelpCtx);
end;

procedure TSqlitePassDatabaseError.Clear;
begin
  FErrorStringList.Clear;
  FErrorCount := 0;
end;

