{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    2006 - 2009

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 05.01.2009

  --------------------------------------------------------------------------- }


{ TSqlitePassRecordset}

constructor TSqlitePassRecordset.Create(Dataset: TSqlitePassDataset);
begin
  Create(Dataset.Database);
  FDataset := Dataset;
  FRecordsCapacity := FDataset.RecordsCacheCapacity;
  FPostPoned := (FDataset.FWriteMode = wmPostponed);
end;

constructor TSqlitePassRecordset.Create(Database: TSqlitePassDatabase);
begin
  inherited Create;
  FDataset := nil;
  FDatabase := Database;
  FActive := False;
  FPostponed := False;
  { Minimum Capacity is set to 100 records
  using a higher value when dealing with very large tables could improve speed }
  FRecordsCapacity := 100;
  {FRecordsMemoryBloc will be allocated in the GetRecords procedure, once we know
   the record size }
  FStringsCapacity := FRecordsCapacity;
  FRecycledStrings := TList.Create;
  FMemoryStreams := TList.Create;
  FRecycledMemoryStreams := TList.Create;
  FCurrentRecordOldValues := nil;
  FBlankRecordInitialization := nil;
  FCalcFieldsNullValuesInitBuffer := nil;
  FUnModifiedRecordInitialization := nil;
end;

destructor TSqlitePassRecordset.Destroy;
begin
  Close;
  FMemoryStreams.Free;
  FRecycledMemoryStreams.Free;
  FRecycledStrings.Free;
  FDataset := nil;
  FDatabase := nil;
  inherited Destroy;
end;


{ --- Opening / Closing --- }

Procedure TSqlitePassRecordset.Open(Sql: String; TablesCount: Word = 0);
begin
Close;
Try
 FActive := True;
 FRowidTablesCount := TablesCount;
 GetRecords(Sql);
Except
 On EAbort do Close; { Safe call to free memory if needed }
    Else begin
         Close; { Safe call to free memory if needed }
         FDatabase.DatabaseError.RaiseExceptionFmt(Msg3013, [CRLF, Sql], -1, FDatabase);
         end;
end;
end;

procedure TSqlitePassRecordset.Close;
begin
if FActive then
   Try
   Reset;
   finally
   FActive := False;
   end;
end;

procedure TSqlitePassRecordset.Reset;
var
i: Integer;
begin
   FCurrentRecordIndex := -1;
   FreeMem(FCurrentRecordOldValues);
   FCurrentRecordOldValues := nil;
   { Clear FRecordsMemoryBlocs }
   For i := 0 to Pred(FRecordsMemoryBlocsCount)
       do FreeMem(FRecordsMemoryBlocs[i]);
   Finalize(FRecordsMemoryBlocs);
   FRecordsMemoryBlocsCount := 0;
   { Clear Records Pointers }
   Finalize(FRecords);
   FRecordsCount := 0;
   FRecordsUsedCapacity := FRecordsCapacity;
   { Blank Record Buffer }
   FreeMem(FBlankRecordInitialization);
   FBlankRecordInitialization := nil;
   { UnModified Value Buffer }
   FreeMem(FUnModifiedRecordInitialization);
   FUnModifiedRecordInitialization := nil;
   { Clear Strings Pointers }
   Finalize(FStrings);
   FStringsCount := 0;
   FStringsCapacityLimit := 0;
   FRecycledStrings.Clear;
   { Clear Blobs and clear Blobs List pointer }
   For i := 0 to Pred(FMemoryStreams.Count)
       do TMemoryStream(FMemoryStreams[i]).Free;
   FMemoryStreams.Clear;
   FRecycledMemoryStreams.Clear;
   { Calc and Lookup null values}
   FreeMem(FCalcFieldsNullValuesInitBuffer);
   FCalcFieldsNullValuesInitBuffer := nil;
   {  }
   Finalize(FInternalFieldsOffset);
end;

function TSqlitePassRecordset.GetFRecordsCount: Integer;
begin
 Result := FRecordsCount;
end;

{ Result : the memory size needed to store a given Datatype in the internal record buffer }
function TSqlitePassRecordset.GetFieldSize(DataType: TFieldType): Integer;
begin
  case DataType of
       ftInteger, ftSmallInt, ftWord, ftDate, ftTime,
       ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString: Result := SizeOf(Integer);
       ftDateTime, ftFloat, ftBCD, ftCurrency: Result := SizeOf(Double);
       ftBoolean: Result := SizeOf(WordBool);
       ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary: Result := SizeOf(Pointer);
       ftAutoInc, ftLargeint: Result := SizeOf(Int64);
       else Result := SizeOf(PChar);
       end;
end;

{ Calculates the size needed to store fields values, including calculated and lookup fields
  and others information in a record
  Returns the number of fields }
function TSqlitePassRecordset.SetFRecordSize: Integer;
var
i, j: Integer;
begin
  FCalcFieldsCount := 0;
  SetLength(FInternalFieldsOffset, 0);

  FBookMarkDataOffset := SizeOf(TBookMarkFlag);
  FRecordStateOffset := FBookMarkDataOffset + SizeOf(Cardinal);
  FRowIdsOffset :=  FRecordStateOffset + SizeOf(TSqlitePassRecordState);
  FFieldsValuesOffset :=  FRowIdsOffset + (FRowidTablesCount * SizeOf(Int64));
  FRecordSize := FFieldsValuesOffset;

  if Assigned(FDataset)
     then begin
          { Set the Fields offset to the maximum length. We will downsize it later }
          SetLength(FInternalFieldsOffset, FDataset.FieldDefs.Count + FDataset.Fields.Count);

          { fkData Fields }
          for i := 0 to Pred(FDataset.FieldDefs.Count)
              do begin
                 FInternalFieldsOffset[i] := FRecordSize;
                 Inc(FRecordSize, GetFieldSize(FDataset.FieldDefs[i].DataType));
                 end;
          { Other fields Types (Calculated or lookup) }
          j := FDataset.FieldDefs.Count;
          for i := 0 to Pred(FDataset.Fields.Count)
              do if FDataset.Fields[i].FieldKind <> fkData then
                 begin
                 FInternalFieldsOffset[j] := FRecordSize;
                 Inc(FRecordSize, GetFieldSize(FDataset.Fields[i].DataType));
                 Inc(j);
                 end;
          { Returns the number of fields, including calc or lookup fields }
          Result := j;
          { Initialise a memory bloc with null values. It will be used to Clear
            the Calc or Lookup fields content }
          FCalcFieldsCount := j-FDataset.FieldDefs.Count;
          GetMem(FCalcFieldsNullValuesInitBuffer, FCalcFieldsCount);
          for i := 0 to Pred(FCalcFieldsCount)
              do PByteBool(FCalcFieldsNullValuesInitBuffer + (i * SizeOf(ByteBool)))^ := True;
          { Set the definitive Fields Offset array size }
          SetLength(FInternalFieldsOffset, j);
         end
     { No TDataset.fields are defined so we use Strings to store any values (stored in FStrings array) }
     else begin
          SetLength(FInternalFieldsOffset, FDataFieldsCount);
          for i := 0 to Pred(FDataFieldsCount) do
              begin
              Inc(FRecordSize, SizeOf(Integer));
              FInternalFieldsOffset[i] := FRecordSize;
              end;
          Result := FDataFieldsCount;
          end;

  { Null values boolean flag buffer }
  FFieldsNullValuesOffset := FRecordSize;
  Inc(FRecordSize, SizeOf(ByteBool)* Result);

  { Modified values boolean flag buffer }
  FFieldsModifiedValuesOffset := FRecordSize;
  GetMem(FUnModifiedRecordInitialization, SizeOf(ByteBool)* Result);
  Inc(FRecordSize, SizeOf(ByteBool)* Result);

  { Prepare a memory bloc with 'blank' values. This memory bloc will be used to initialize
    each new record in the AddRecord function }
  GetMem(FBlankRecordInitialization, FRecordSize);
  { Set Record state }
  TSqlitePassRecordState((FBlankRecordInitialization + FRecordStateOffset)^) := rsUnModified;
  { Initialize Null values Flag and Modified Flags }
  for j := 0 to Pred(Result) do
      begin
      i := j * SizeOf(ByteBool);
      PByteBool(FBlankRecordInitialization + FFieldsNullValuesOffset + i)^ := True;
      PByteBool(FBlankRecordInitialization + FFieldsModifiedValuesOffset + i)^ := False;
      PByteBool(FUnModifiedRecordInitialization + i)^ := False;
      end;
end;


{ --- Gets the records matching the SQL statement from sqlite engine --- }

procedure TSqlitePassRecordset.GetRecords(Sql: String);
var
Accept: Boolean;
PSqliteData: Pointer;
ColumnCount, ColumnIndex: Cardinal;
Buffer: PChar;
ValueOffset, BufferSize, ErrorCode,LowerLimit,
UpperLimit, LimitIdx, BufferFieldPos: Integer;
RecordIndex: Integer;
begin
Try
PSqliteData := nil;
Reset;
FDatabase.Engine.PrepareStmt(PSqliteData, Sql);
ColumnCount := SqliteDbv3_column_count(PSqliteData);

{ Sometimes there is no data at all }
if ColumnCount = 0
   then Abort;

FDataFieldsCount := ColumnCount - FRowIdTablesCount;
{ Set the Lower and Upper Limit to retrieve only the rows requested }
LimitIdx := 0;
LowerLimit := 1;
UpperLimit := Pred(MaxLongInt);

FFieldsCount := SetFRecordSize;
ValueOffset := FRowIdsOffset;

if Assigned(FDataset) then
   begin
   if FDataset.Filtered then
      begin
      LowerLimit := Max(1, FDataset.FilterRecordLowerLimit);
      if FDataset.FilterRecordUpperLimit <> 0
         then UpperLimit := Succ(FDataset.FilterRecordUpperLimit);
      end;
  InitInternalFieldsInfos(FDataset.Fields);
  end;

{ We step through recordset until we reach LowerLimit or we get an error }
Repeat
  ErrorCode :=SqliteDbv3_Step(PSqliteData);
  Inc(LimitIdx);
until (ErrorCode <> SQLITE_ROW) or (LimitIdx = LowerLimit);

{ Proceed with every records until we reach UpperLimit or we get an error }
While (ErrorCode = SQLITE_ROW) and (LimitIdx < UpperLimit) do
  begin
  { Update Records Count and Capacity }
  RecordIndex := AddRecord;
  Buffer := FRecords[RecordIndex];

  { Move the pointer to the first field buffer address }
  Inc(Buffer, ValueOffset);

  { We get the fields values for the current data row }
  for ColumnIndex := 0 to Pred(ColumnCount) do
      begin
      if Assigned(FDataset) then
         { fields are defined }
         begin
         BufferFieldPos := ColumnIndex-FRowIdTablesCount;
         if ColumnIndex < FRowIdTablesCount
            { Stores the rowid for each table returned by SQL }
            then BufferSize := SqliteValueToBuffer(PSqliteData, ColumnIndex, BufferFieldPos, ftLargeInt, Buffer)
            { Stores the fields values }
            else BufferSize := SqliteValueToBuffer(PSqliteData, ColumnIndex, BufferFieldPos, FDataset.FieldDefs[BufferFieldPos].DataType, Buffer);
         Inc(Buffer, BufferSize);
         end
         { No fields are defined, so we store all the values as string }
         else begin
              AddString(Buffer, SqliteDbv3_column_text(PSqliteData,ColumnIndex));
              SetFieldNullValue(RecordIndex, ColumnIndex, False);
              Inc(Buffer, SizeOf(Integer));
              end;
      end; { for }

  Inc(LimitIdx);

  { If needed, we send filter event now }
  if Assigned(FDataset) then
     begin
     { dsFilter is set only when Dataset is Filtered and OnFilterRecord is Assigned }
     if FDataset.State = dsFilter then
        begin
        CurrentRecordIndex := Pred(FRecordsCount);
        Accept := True;
        FDataset.OnFilterRecord(FDataset, Accept);
        If Not Accept then
           begin
           DeleteRecord(Pred(FRecordsCount));
           Dec(LimitIdx);
           end;
        end;
     end;

  ErrorCode :=SqliteDbv3_Step(PSqliteData);
end; { While }

// TODO (FStatementTimeout=0) or ((FStatementTimeout>0) and (integer(GetTickCount)>=t));

{ Move to the first record }
FCurrentRecordIndex := -1;
FOldCurrentRecordIndex := -1;

{ Create a utility record to be able to restore a record modified by user to
  it previous state if needed }
FCurrentRecordOldValues := AllocMem(FRecordSize);
finally
FDatabase.Engine.UnPrepareStmt(PSqliteData);
end;
end;

{ Translates a SQLite value to its internal storage value in memory
  Returns the size of the value }
function TSqlitePassRecordset.SqliteValueToBuffer(Const PSqliteData: Pointer; Const ColumnIndex: Integer; Const BufferFieldPos: Integer; Const DataType: TFieldType; Buffer: PChar): Integer;
var
IntegerValue: Integer;
BooleanValue: WordBool;
DoubleValue: Double;
Int64Value: Int64;
BlobStream: TMemoryStream;
BlobSize: Cardinal;
begin
Result := 0;
Try
 Case DataType of
  ftSmallint, ftInteger, ftWord:
    begin
    Result := SizeOf(Integer);
    if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
    IntegerValue := SqliteDbv3_column_int(PSqliteData,ColumnIndex);
    Move(IntegerValue, Buffer^, SizeOf(Integer));
    end;

  ftBoolean:
    begin
    Result := SizeOf(WordBool);
    if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
    Case FDatabase.DatatypeOptions.FBooleanStorage of
         asInteger : BooleanValue := SqliteDbv3_column_int(PSqliteData,ColumnIndex) <> 0;
         asText    : FDatabase.FTranslator.BooleanTextToWordBool(SqliteDbv3_column_text(PSqliteData,ColumnIndex),BooleanValue);
         end;
    Move(BooleanValue, Buffer^, SizeOf(WordBool));
    end;

  ftFloat, ftCurrency, ftBCD:
    begin
    Result := SizeOf(Double);
    if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
    DoubleValue := SqliteDbv3_column_double(PSqliteData,ColumnIndex);
    Move(DoubleValue, Buffer^, SizeOf(Double));
    end;

  ftDate:
    begin
    Result := SizeOf(Integer);
    if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
    Case FDatabase.DatatypeOptions.DateStorage of
         asInteger : IntegerValue := SqliteDbv3_column_int(PSqliteData,ColumnIndex);
         asText    : FDatabase.FTranslator.DateTextToInteger(SqliteDbv3_column_text(PSqliteData,ColumnIndex),IntegerValue);
         end;
    Move(IntegerValue, Buffer^, SizeOf(Integer));
    end;

  ftTime:
    begin
    Result := SizeOf(Integer);
    if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
    Case FDatabase.DatatypeOptions.TimeStorage of
         asInteger : IntegerValue := SqliteDbv3_column_int(PSqliteData,ColumnIndex);
         asText    : FDatabase.FTranslator.TimeTextToInteger(SqliteDbv3_column_text(PSqliteData,ColumnIndex),IntegerValue);
         end;
    Move(IntegerValue, Buffer^, SizeOf(Integer));
    end;

  ftDateTime:
    begin
    Result := SizeOf(Double);
    if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
    Case FDatabase.DatatypeOptions.DateTimeStorage of
         dtsDateTime : SqliteDbv3_column_double(PSqliteData,ColumnIndex);
         dtsJulian   : FDatabase.FTranslator.JulianDateTimeToDouble(SqliteDbv3_column_double(PSqliteData,ColumnIndex),DoubleValue);
         dtsText     : FDatabase.FTranslator.DateTimeTextToDouble(SqliteDbv3_column_text(PSqliteData,ColumnIndex),DoubleValue);
         dtsUnix     : FDatabase.FTranslator.UnixDateTimeToDouble(SqliteDbv3_column_int64(PSqliteData,ColumnIndex),DoubleValue);
         dtsMac      : FDatabase.FTranslator.MacDateTimeToDouble(SqliteDbv3_column_int64(PSqliteData,ColumnIndex),DoubleValue);
         end;
    Move(DoubleValue, Buffer^, SizeOf(Double));
    end;

 ftAutoInc, ftLargeint:
   begin
   Result := SizeOf(Int64);
   if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
   Int64Value := SqliteDbv3_column_int64(PSqliteData,ColumnIndex);
   Move(Int64Value, Buffer^, SizeOf(Int64));
   end;

 ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString:
   begin
   Result := SizeOf(Integer);
   if SqliteDbv3_column_type(PSqliteData, ColumnIndex) = SQLITE_NULL then Exit;
   AddString(Buffer, SqliteDbv3_column_text(PSqliteData, ColumnIndex));
   end;

 ftBlob, ftGraphic, ftMemo, ftFmtMemo:
   begin
   Result := SizeOf(Integer);
   BlobSize := SqliteDbv3_column_bytes(PSqliteData, ColumnIndex);
   if BlobSize = 0 then Exit;
   BlobStream := TMemoryStream(FMemoryStreams[AddMemoryStream(Buffer)]);
   BlobStream.Write(SqliteDbv3_column_blob(PSqliteData, ColumnIndex)^, BlobSize);
   end;
 end; { Case}

{ By default, value is already initialized to Null in record buffer (AddRecord takes care of this)
  If we succeeded then we set the field value as NOT Null }
  if BufferFieldPos > -1
     then SetFieldNullValue(Pred(FRecordsCount), BufferFieldPos, False);
Except
{ Exception while retreiving data are trapped here
  so we can continu with the next available value }
On E: Exception do;
end;
end;

{ --- Navigating --- }

Procedure TSqlitePassRecordset.First;
begin
CurrentRecordIndex := -1;
end;

Procedure TSqlitePassRecordset.Prior;
begin
MoveBy(-1)
end;

Procedure TSqlitePassRecordset.Next;
begin
MoveBy(1)
end;

Procedure TSqlitePassRecordset.Last;
begin
CurrentRecordIndex := FRecordsCount;
{$IFDEF FPC}
FOldCurrentRecordIndex := FRecordsCount;
{$ENDIF}
end;

Procedure TSqlitePassRecordset.MoveBy(Offset: Integer);
begin
Try
MoveState := grOk;
Case Offset of
   { Move to Previous record }
   -1 : begin
        Dec(FCurrentRecordIndex);
        if FCurrentRecordIndex < 0 then
           begin
           MoveState := grBOF;
           FCurrentRecordIndex := -1;
           end;
        end;
   { Current record }
    0 : if (FCurrentRecordIndex < 0) or (FCurrentRecordIndex > Pred(FRecordsCount))
           then MoveState := grError;

   { Move to the Next record }
    1 : begin
        Inc(FCurrentRecordIndex);
        if (FCurrentRecordIndex > Pred(FRecordsCount)) then
           begin
           MoveState := grEOF;
           FCurrentRecordIndex := FRecordsCount;
           end;
        end;
    end;
if (FCurrentRecordIndex > -1) and (FCurrentRecordIndex < FRecordsCount)
   then SetBookMarkFlag(FCurrentRecordIndex, bfCurrent);

Except
MoveState := grError
end;
end;


{ --- Managing Records (Create a new record, Add, Insert, Delete from records list) --- }

 { Set default values if needed }
(*For i := 0 to Pred(Fields.Count) do
    if Fields[i].FieldKind = fkData
       then Result^.FieldsValues[Pred(Fields[i].FieldNo)] := Fields[i].DefaultExpression;*)

{ Add a new 'temporary' record and upsize the memory if necessary
  Returns the index of the next available record
  We don't inc FRecordCount so Dataset doesn't know about the new record
  until the PostRecord procedure is called  }
function TSqlitePassRecordset.AddTempRecord: Integer;
begin

if FRecordsUsedCapacity = FRecordsCapacity then
   begin
   Inc(FRecordsMemoryBlocsCount);
   SetLength(FRecords, FRecordsCapacity * FRecordsMemoryBlocsCount);
   SetLength(FRecordsMemoryBlocs, FRecordsMemoryBlocsCount);
   { AllocMem is used instead of GetMem because it initializes memory with #0 }
   FRecordsMemoryBlocs[Pred(FRecordsMemoryBlocsCount)] := AllocMem(FRecordSize * FRecordsCapacity);
   FRecordsUsedCapacity := 0;
   end;
FRecords[FRecordsCount] := FRecordsMemoryBlocs[Pred(FRecordsMemoryBlocsCount)] + (FRecordSize * FRecordsUsedCapacity);
{ Initialize the new record with 'blank' values }
Move(FBlankRecordInitialization^, FRecords[FRecordsCount]^, FRecordSize);
Result := FRecordsCount;
end;

{ Add a new record }
function TSqlitePassRecordset.AddRecord: Integer;
begin
  Result := AddTempRecord;
  Inc(FRecordsCount);
  Inc(FRecordsUsedCapacity);
end;

{ Insert a new record :
  The record itself is added to the next available place
  in the last memory bloc.
  Then a new pointer is assigned to this record and inserted a the right place in the FRecords array }
Procedure TSqlitePassRecordset.InsertRecord(Index: Integer);
var
TempRecord: PChar;
begin
 TempRecord := FRecords[FRecordsCount];
 System.Move(FRecords[Index], FRecords[Index + 1],(FRecordsCount - Index) * SizeOf(PChar));
 FRecords[Index] := TempRecord;
end;

{ Make the record added with AddTempRecord 'visible' for the dataset }
procedure TSqlitePassRecordset.PostRecord;
begin
If (FOldCurrentRecordIndex > -1) and (FOldCurrentRecordIndex < FRecordsCount)
   then begin
   InsertRecord(FOldCurrentRecordIndex);
   FCurrentRecordIndex := FOldCurrentRecordIndex;
   FOldCurrentRecordIndex := -1;
   end;
  Inc(FRecordsCount);
  Inc(FRecordsUsedCapacity);
end;

procedure TSqlitePassRecordset.Cancel;
begin
{ Move back to the Old selected record }
if FOldCurrentRecordIndex > -1 then
   begin
   FCurrentRecordIndex := FOldCurrentRecordIndex;
   FOldCurrentRecordIndex := -1;
   end;
end;

{ Add a String and Returns the Index of the new string in Fstrings Array }
function TSqlitePassRecordset.AddString(Buffer: PChar; StrValue: String): Integer;
begin
 if FRecycledStrings.Count > 0
    then begin
    Result := PInteger(FRecycledStrings[Pred(FRecycledStrings.Count)])^;
    FRecycledStrings.Delete(Pred(FRecycledStrings.Count));
    end
    else begin
    Result := FStringsCount;
    { Resizes the strings array if capacity is over }
    if FStringsCount = FStringsCapacityLimit then
       begin
       Inc(FStringsCapacityLimit, FStringsCapacity);
       SetLength(FStrings, FStringsCapacityLimit);
       end;
    Inc(FStringsCount);
    end;
  FStrings[Result] := StrValue;
  Move(Result, Buffer^, SizeOf(Integer));
 end;

function TSqlitePassRecordset.AddString(RecordIndex, FieldIndex: Integer; StrValue: String): Integer;
var
Buffer: PChar;

begin
 Buffer := GetFieldBuffer(RecordIndex, FieldIndex);
 Result := AddString(Buffer, StrValue);
 SetFieldNullValue(RecordIndex, FieldIndex, False);
end;

procedure TSqlitePassRecordset.ClearString(RecordIndex, FieldIndex: Integer);
var
StrIndex: Integer;
begin
if Not FieldIsNull(RecordIndex, FieldIndex) then
   begin
   StrIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
   FStrings[StrIndex] := '';
   FRecycledStrings.Add(Pointer(StrIndex));
   SetFieldNullValue(RecordIndex, FieldIndex, True);
   end;
end;

{ Add a MemoryStream and Returns the index of the MemoryStream in FMemoryStreams }
function TSqlitePassRecordset.AddMemoryStream(Buffer: PChar): Integer;
begin
  if FRecycledMemoryStreams.Count > 0
     then begin
     Result := PInteger(FRecycledMemoryStreams[Pred(FRecycledMemoryStreams.Count)])^;
     FRecycledMemoryStreams.Delete(Pred(FRecycledMemoryStreams.Count));
     end
     else Result := FMemoryStreams.Add(TMemoryStream.Create);
  Move(Result, Buffer^, SizeOf(Integer));
end;

function TSqlitePassRecordset.AddMemoryStream(RecordIndex, FieldIndex: Integer): Integer;
var
Buffer: PChar;
begin
 Buffer := GetFieldBuffer(RecordIndex, FieldIndex);
 Result := AddMemoryStream(Buffer);
 SetFieldNullValue(RecordIndex, FieldIndex, False);
end;

procedure TSqlitePassRecordset.ClearMemoryStream(RecordIndex, FieldIndex: Integer);
var
BlobIndex: Integer;
begin
if Not FieldIsNull(RecordIndex, FieldIndex) then
   begin
   BlobIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
   TMemoryStream(FMemoryStreams[BlobIndex]).Clear;
   FRecycledMemoryStreams.Add(Pointer(BlobIndex));
   SetFieldNullValue(RecordIndex, FieldIndex, True);
   end;
end;

Function TSqlitePassRecordset.GetFieldAsBlob(RecordIndex, FieldIndex: Integer): TMemoryStream;
var
BlobIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then Result := nil
   else begin
        BlobIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        Result := FMemoryStreams[BlobIndex];
        end;
end;


{ Delete FRecords[Index] from the FRecords array and downsize the array
  The record itself is not freed
  NO checking of array bounds }
procedure TSqlitePassRecordset.DeleteRecord(Index: Integer);
begin
  ClearRecord(Index);
  Dec(FRecordsCount);
  Dec(FRecordsUsedCapacity);
  { If Index = FRecordsCount, we don't have to move the memory }
  if Index < FRecordsCount then
     begin
     System.Move(FRecords[Succ(Index)], FRecords[Index],(FRecordsCount - Index) * SizeOf(PChar));
     CurrentRecordIndex := Index;
     end
     else CurrentRecordIndex := FRecordsCount;
end;


procedure TSqlitePassRecordset.CopyRecord(Source, Dest: PChar);
var
i, StrIndex, SourceBlobStreamIndex, DestBlobStreamIndex: Integer;
SourceString: String;
SourceBuffer, DestBuffer: PChar;
begin
{ We should not need to copy record if FDataset is not assigned, so it is not
  implemented }
if Not Assigned(FDataset) then Exit;
Move(Source^, Dest^, FRecordSize);
(*For i := 0 to Pred(FDataset.Fields.Count) do
   begin
   if (FDataset.Fields[i].DataType in SqlitePassBlobFieldTypes) then
       { Copy Blobs }
       begin
       SourceBuffer := Source + FFieldsOffset[FDataset.Fields[i].Index];
       SourceBlobStreamIndex := PInteger(SourceBuffer)^;
       if SourceBlobStreamIndex > -1 then
          begin
          DestBlobStreamIndex := AddMemoryStream(Dest + FFieldsOffset[FDataset.Fields[i].Index]);
          TMemoryStream(FMemoryStreams[DestBlobStreamIndex]).LoadFromStream(TMemoryStream(FMemoryStreams[SourceBlobStreamIndex]));
          end;
      end
      { Copy Strings }
      else if (FDataset.Fields[i].DataType in SqlitePassStringFieldTypes) then
           begin
           StrIndex :=  PInteger(Source + FFieldsOffset[FDataset.Fields[i].Index])^;
           SourceString := FStrings[StrIndex];
           AddString(Dest + FFieldsOffset[FDataset.Fields[i].Index], SourceString);
           end;
     end;   *)
end; { for }

{ SqlitePassBlobFieldTypes = [ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary];
  SqlitePassStringFieldTypes = [ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString]; }
procedure TSqlitePassRecordset.ClearRecord(Index: Integer);
var
i: Integer;
begin
For i := 0 to Pred(FDataset.Fields.Count) do
   begin
   Case FDataset.Fields[i].DataType of
        { Clear Blobs }
        ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftTypedBinary: ClearMemoryStream(Index, FDataset.Fields[i].Tag);
        { Clear Strings content }
        ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString: ClearString(Index, FDataset.Fields[i].Tag);
        end;
   end; { for }
{ Clear Buffer }
Move(FBlankRecordInitialization^, FRecords[Index]^, FRecordSize);
end;

Procedure TSqlitePassRecordset.ExchangeRecords(RecordIndex1, RecordIndex2: Integer);
var
TempRecord: Pchar;
begin
 TempRecord := FRecords[RecordIndex1];
 FRecords[RecordIndex1] := FRecords[RecordIndex2];
 FRecords[RecordIndex2] := TempRecord;
end;


{ --- Current record old values backup / restore when editing --- }

procedure TSqlitePassRecordset.SaveCurrentRecordOldValues;
begin
 CopyRecord(FRecords[FCurrentRecordIndex], FCurrentRecordOldValues);
end;

procedure TSqlitePassRecordset.RestoreCurrentRecordOldValues;
begin
 CopyRecord(FCurrentRecordOldValues, FRecords[FCurrentRecordIndex]);
end;


{ Read / Write Record BookMarkFlag }
Function TSqlitePassRecordset.GetBookMarkFlag(RecordIndex: Integer): TBookMarkFlag;
begin
  Result := TBookMarkFlag(FRecords[RecordIndex]^);
end;

Procedure TSqlitePassRecordset.SetBookMarkFlag(RecordIndex: Integer; Value: TBookMarkFlag);
begin
{$IFDEF FPC}
  Move(Value, FRecords[RecordIndex]^, SizeOf(Value));
{$ELSE}
  TBookMarkFlag(FRecords[RecordIndex]^) := Value;
{$ENDIF}
end;

{ Read / Write a record BookMarkData }
Function TSqlitePassRecordset.GetBookMarkData(RecordIndex: Integer): Cardinal;
begin
  Result := PCardinal(FRecords[RecordIndex]+ FBookMarkDataOffset)^;
  if FOldCurrentRecordIndex = -1
     then FOldCurrentRecordIndex := RecordIndex;
end;

Procedure TSqlitePassRecordset.SetBookMarkData(RecordIndex: Integer; Value: Cardinal);
begin
  PCardinal(FRecords[RecordIndex]+ FBookMarkDataOffset)^ := Value;
end;

{ Read / Write Record RecordState }
Function TSqlitePassRecordset.GetRecordState(RecordIndex: Integer):  TSqlitePassRecordState;
begin
  Result := TSqlitePassRecordState((FRecords[RecordIndex] + FRecordStateOffset)^)
end;

Procedure TSqlitePassRecordset.SetRecordState(RecordIndex: Integer; Value: TSqlitePassRecordState);
var
PreviousRecordState: TSqlitePassRecordState;

begin
if FPostPoned then
  begin
  PreviousRecordState := GetRecordState(RecordIndex);
  Case PreviousRecordState of
       { rsUnModified can be changed to any other state }
       rsUnModified : begin
                      TSqlitePassRecordState((FRecords[RecordIndex] + FRecordStateOffset)^) := Value;
                      if Value = rsUnModified
                         then Move(FUnModifiedRecordInitialization^,
                                   (FRecords[RecordIndex] + FFieldsModifiedValuesOffset)^,
                                   FFieldsCount);
                      end;
       { rsModified can be changed only to rsDeleted }
       rsModified   : if Value = rsDeleted
                         then TSqlitePassRecordState((FRecords[RecordIndex] + FRecordStateOffset)^) := Value;

       { rsModified can be changed only to rsDeleted. In this case, record is deleted from Memory }
       rsInserted   : if Value = rsDeleted
                         then DeleteRecord(RecordIndex);
       end;
       { rsDeleted  cannot be changed to another value }
  end;
end;

{ Read / Write Record Rowid - TableNo is zero indexed }
Function TSqlitePassRecordset.GetRowId(RecordIndex: Integer; TableNo: Word): Int64;
begin
  if FRowidTablesCount > 0
     then Result := PInt64(FRecords[RecordIndex] + FRowIdsOffset + (TableNo * SizeOf(Int64)))^
     else Result := -1;
end;

Procedure TSqlitePassRecordset.SetRowId(RecordIndex: Integer; TableNo: Word; Value: Int64);
begin
  PInt64(FRecords[RecordIndex] + FRowIdsOffset + (TableNo * SizeOf(Int64)))^ := Value;
end;


{ FieldIndex is zero indexed }
Function TSqlitePassRecordset.GetFieldBuffer(RecordIndex, FieldIndex: Integer): PChar;
begin
  Result := FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex];
end;

Function TSqlitePassRecordset.GetFieldAsString(RecordIndex, FieldIndex: Integer): String;
var
StrIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then Result := ''
   else begin
        StrIndex := PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        Result := FStrings[StrIndex];
        end;
end;

Procedure TSqlitePassRecordset.SetFieldAsString(RecordIndex, FieldIndex: Integer; Value: String);
var
StrIndex: Integer;
begin
if FieldIsNull(RecordIndex, FieldIndex)
   then AddString(RecordIndex, FieldIndex, Value)
   else begin
        StrIndex :=  PInteger((FRecords[RecordIndex] + FInternalFieldsOffset[FieldIndex]))^;
        FStrings[StrIndex] := Value;
        end;
end;


Function TSqlitePassRecordset.GetFieldAsVariant(RecordIndex, FieldIndex: Integer; Field: TField): Variant;
var
Buffer: PChar;
begin
Result := UnAssigned;
Buffer := GetFieldBuffer(RecordIndex, FieldIndex);

if FieldIsNull(RecordIndex, FieldIndex)
   then Result := Null
   else
   Case Field.DataType of
        ftWord,
        ftSmallint  : Result := PSmallInt(Buffer)^;
        ftInteger   : Result := PInteger(Buffer)^;
        ftBoolean   : Result := PWordBool(Buffer)^;
        ftFloat,
        ftCurrency,
        ftBCD       : Result := PDouble(Buffer)^;
        ftDate,
        ftTime      : Result := PInteger(Buffer)^;
        ftDateTime  : Result := VarFromDateTime(PDouble(Buffer)^);
        ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString
                    : Result := GetFieldAsString(RecordIndex, FieldIndex);
        end;
end;


{ Read and Write a record 'Null' Boolean Flag }
Function TSqlitePassRecordset.FieldIsNull(RecordIndex, FieldIndex: Integer): Boolean;
begin
  Result := PByteBool(FRecords[RecordIndex] + FFieldsNullValuesOffset + FieldIndex)^ = True;
end;

Procedure TSqlitePassRecordset.SetFieldNullValue(RecordIndex, FieldIndex: Integer; Value: ByteBool);
begin
  PByteBool(FRecords[RecordIndex] + FFieldsNullValuesOffset + FieldIndex)^ := Value;
end;

{ Read and Write a record 'Modified' Boolean Flag }
Function TSqlitePassRecordset.FieldWasModified(RecordIndex, FieldIndex: Integer): Boolean;
begin
  Result := PByteBool(FRecords[RecordIndex] + FFieldsModifiedValuesOffset + FieldIndex)^ = True;
end;

Procedure TSqlitePassRecordset.SetFieldWasModified(RecordIndex, FieldIndex: Integer; Value: ByteBool);
begin
  PByteBool(FRecords[RecordIndex] + FFieldsModifiedValuesOffset + FieldIndex)^ := Value;
end;

{ --- Direct access to fields values, when no TFields are defined ---
  RecordIndex and FieldIndex are zero based }

{ Returns a string for a given RecordIndex and FieldIndex }
Function TSqlitePassRecordset.GetTextItem(RecordIndex, FieldIndex: Integer): String;
var
StrIndex: Integer;
begin
  StrIndex := PInteger(PChar(FRecords[RecordIndex] + FFieldsValuesOffset + (SizeOf(Integer)* FieldIndex)))^;
  Result := FStrings[StrIndex];
end;

{ Set a string for a given RecordIndex and FieldIndex }
Procedure TSqlitePassRecordset.SetTextItem(RecordIndex, FieldIndex: Integer; Value: String);
var
StrIndex: Integer;
begin
  StrIndex := PInteger(PChar(FRecords[RecordIndex] + FFieldsValuesOffset + (SizeOf(Integer)* FieldIndex)))^;
  FStrings[StrIndex] := Value;
end;

{ Add a string for a given RecordIndex and FieldIndex
  Field Index is zero Indexed }
Procedure TSqlitePassRecordset.AddTextItem(RecordIndex, FieldIndex: Integer; Value: String);
var
Buffer: PChar;
begin
  Buffer := PChar(FRecords[RecordIndex] + FFieldsValuesOffset + (SizeOf(Integer)* FieldIndex));
  AddString(Buffer, Value);
  SetFieldNullValue(RecordIndex, FieldIndex, False)
end;


{ Internal information to retrieve a internal field position or a internal field offset
  in buffer from a given TDataset.Field.Index }

procedure TSqlitePassRecordset.InitInternalFieldsInfos(Fields: TFields);
var
i, j: Integer;
begin
FDataset.FHasCalcFields := False;
j := 0;
for i := 0 to Pred(Fields.Count) do
    begin
    Case Fields[i].FieldKind of
         fkData:      Fields[i].Tag := Pred(Fields[i].FieldNo);

         fkCalculated,
         fkLookup:    begin
                      Fields[i].Tag := FDataFieldsCount + j;
                      Inc(j);
                      FDataset.FHasCalcFields := True;
                      end;
        end; { Case }
    end;
end;


{ Set all the Calculated and Lookup fields as Null }
procedure TSqlitePassRecordset.ClearCalcFields(RecordIndex: Integer);
var
CalcFieldsNullValues: PChar;
begin
if Assigned(FCalcFieldsNullValuesInitBuffer) then
   begin
   CalcFieldsNullValues := FRecords[RecordIndex] + FFieldsNullValuesOffset + FDataFieldsCount;
   Move(FCalcFieldsNullValuesInitBuffer, CalcFieldsNullValues, FCalcFieldsCount);
   end;
end;


{ Sorting }

Procedure TSqlitePassRecordset.SortRecords(SortedFields: TList);

  function CompareRecords(RecordIndex1: Integer; RecordBuffer: PChar): Integer;
  var
  i: Integer;
  begin
  Result := 0;
  for i := 0 to Pred(SortedFields.Count) do
      begin
      Result := CompareFieldValues(TSqlitePassSortedField(SortedFields[i]).FieldIndex,
                                   RecordIndex1, RecordBuffer,
                                   TSqlitePassSortedField(SortedFields[i]).Descendant);
      { Skips the others fields comparisons if values are already different }
      if Result <> 0 then Break
      end;
  end;

  Procedure InternalSortRecords(iLo, iHi: Integer);
  var
    Lo, Hi: Integer;
    MidRecordBuffer: PPChar;
  begin
    Lo := iLo;
    Hi := iHi;
    New(MidRecordBuffer);
    MidRecordBuffer^ := FRecords[(Lo + Hi) shr 1];
    repeat
     while CompareRecords(Lo, MidRecordBuffer^) < 0 do Inc(Lo);
     while CompareRecords(Hi, MidRecordBuffer^) > 0 do Dec(Hi);
     if Lo <= Hi then
      begin
        ExchangeRecords(Lo,Hi);
        Inc(Lo);
        Dec(Hi);
      end;
    until Lo > Hi;
    if Hi > iLo then InternalSortRecords(iLo, Hi);
    if Lo < iHi then InternalSortRecords(Lo, iHi);
    Dispose(MidRecordBuffer);
  end;
begin { SortRecords }
  InternalSortRecords(0, Pred(FRecordsCount));
end;


{ Utility function to compare two values based on the same DataType.
  This function is used to sort the recordset }
function TSqlitePassRecordset.CompareFieldValues(Const FieldIndex: Integer; RecordIndex1: Integer; RecordBuffer: PChar; Descending: Boolean = False): Integer;
var
FieldPos: Integer;
Buffer1, Buffer2: PChar;
Buffer1IsNull, Buffer2IsNull: Boolean;
MemoryStream1, MemoryStream2: TMemorySTream;
MemoText1, MemoText2: String;
DataType: TFieldType;
begin
Result := 0;
FieldPos := FDataset.Fields[FieldIndex].Tag;
Buffer1IsNull := FieldIsNull(RecordIndex1, FieldPos);
Buffer2IsNull := (PByteBool(RecordBuffer + FFieldsNullValuesOffset + FieldPos)^ = True);

if Buffer1IsNull and Not Buffer2IsNull then Result := -1
   else if Buffer2IsNull and Not Buffer1IsNull then Result := 1
        { The two values are NOT Null }
        else if Not (Buffer1IsNull and Buffer2IsNull) then
                begin
                Buffer1 := GetFieldBuffer(RecordIndex1, FieldPos);
                Buffer2 := RecordBuffer + FInternalFieldsOffset[FieldPos];
                DataType := FDataset.Fields[FieldIndex].DataType;

                Case DataType of
                      ftBoolean :
                      if PWordBool(Buffer1)^ > PWordBool(Buffer2)^ then Result := 1
                         else if PWordBool(Buffer1)^ < PWordBool(Buffer2)^ then Result := -1;

                      ftWord, ftSmallint, ftInteger, ftDate, ftTime :
                      if PInteger(Buffer1)^ > PInteger(Buffer2)^ then Result := 1
                         else if PInteger(Buffer1)^ < PInteger(Buffer2)^ then Result := -1;

                      ftAutoInc, ftLargeInt :
                      if PInt64(Buffer1)^ > PInt64(Buffer2)^ then Result := 1
                         else if PInt64(Buffer1)^ < PInt64(Buffer2)^ then Result := -1;

                      ftBCD, ftCurrency, ftFloat, ftDateTime :
                      if PDouble(Buffer1)^ > PDouble(Buffer2)^ then Result := 1
                         else if PDouble(Buffer1)^ < PDouble(Buffer2)^ then Result := -1;

                      ftString, ftBytes, ftVarBytes, ftFixedChar, ftWideString :
                      Result := AnsiCompareStr(FStrings[PInteger(Buffer1)^], FStrings[PInteger(Buffer2)^]);

                      ftMemo, ftFmtMemo :
                      begin
                      MemoryStream1 := FMemoryStreams[PInteger(Buffer1)^];
                      MemoryStream2 := FMemoryStreams[PInteger(Buffer2)^];
                      if MemoryStream1.Size > 0
                         then SetString(MemoText1, PChar(MemoryStream1.Memory), MemoryStream1.Size)
                         else MemoText1 := '';

                      if MemoryStream2.Size > 0
                         then SetString(MemoText2, PChar(MemoryStream2.Memory), MemoryStream2.Size)
                         else MemoText2 := '';

                      Result := AnsiCompareStr(MemoText1, MemoText2);
                      end;
                   end; { Case }
                end; { if Not (Buffer1IsNull and Buffer2IsNull) }

{ Inverts the result if Descending Order }
If Descending and (Result <> 0)
   then Result := Result * -1;
end;

{ --- Clear Postponed Records Changes --- }
Procedure TSqlitePassRecordset.ClearModifiedRecords;
begin
  SetLength(FModifiedRecords, 0);
end;


